
#ifndef SimdMATRIX4X4_H
#define SimdMATRIX4X4_H


#include "SimdVector4.h"
#include "SimdTransform.h"

// Row-major 4x4 matrix

class SimdMatrix4x4 {
public:
	/**
	 * Empty contructor.
	 */
    SimdMatrix4x4() {}
	/**
	 * Initialize all fields with the values pointed at by m. A
	 * contigous block of 16 values is read.  */
    SimdMatrix4x4(const float *m) { setValue(m); }
	/**
	 * Initialize all fields with the values pointed at by m. A
	 * contigous block of 16 values is read.  */
    SimdMatrix4x4(const double *m) { setValue(m); }
    
	/**
	 * Initialise with these 16 explicit values.
	 */
    SimdMatrix4x4(SimdScalar xx, SimdScalar xy, SimdScalar xz, SimdScalar xw,
                 SimdScalar yx, SimdScalar yy, SimdScalar yz, SimdScalar yw,
                 SimdScalar zx, SimdScalar zy, SimdScalar zz, SimdScalar zw,
                 SimdScalar wx, SimdScalar wy, SimdScalar wz, SimdScalar ww) { 
        setValue(xx, xy, xz, xw, 
                 yx, yy, yz, yw,
                 zx, zy, zz, zw,
				 wx, wy, wz, ww);
    }
	
	/** 
	 * Initialize from an SimdTransform.
	 */ 
	SimdMatrix4x4(const SimdTransform &t) {

		const SimdMatrix3x3 &basis = t.getBasis();
		const SimdVector3 &origin = t.getOrigin();  	

		setValue(
			basis[0][0],basis[0][1],basis[0][2],origin[0],
			basis[1][0],basis[1][1],basis[1][2],origin[1],
			basis[2][0],basis[2][1],basis[2][2],origin[2],
			SimdScalar(0),SimdScalar(0),SimdScalar(0),SimdScalar(1)
		);
	}
		
	/**
	 * Get the i-th row.
	 */
    SimdVector4&       operator[](int i)       { return m_el[i]; }
	/**
	 * Get the i-th row.
	 */
    const SimdVector4& operator[](int i) const { return m_el[i]; }

    /**
	 * Set the matrix to the values pointer at by m. A contiguous
	 * block of 16 values is copied.  */
    void setValue(const float *m) {
        m_el[0][0] = *m++; m_el[1][0] = *m++; m_el[2][0] = *m++; m_el[3][0] = *m++;
        m_el[0][1] = *m++; m_el[1][1] = *m++; m_el[2][1] = *m++; m_el[3][1] = *m++;
        m_el[0][2] = *m++; m_el[1][2] = *m++; m_el[2][2] = *m++; m_el[3][2] = *m++;
        m_el[0][3] = *m++; m_el[1][3] = *m++; m_el[2][3] = *m++; m_el[3][3] = *m;
    }

    /**
	 * Set the matrix to the values pointer at by m. A contiguous
	 * block of 16 values is copied.
	 */
    void setValue(const double *m) {
        m_el[0][0] = *m++; m_el[1][0] = *m++; m_el[2][0] = *m++; m_el[3][0] = *m++;
        m_el[0][1] = *m++; m_el[1][1] = *m++; m_el[2][1] = *m++; m_el[3][1] = *m++;
        m_el[0][2] = *m++; m_el[1][2] = *m++; m_el[2][2] = *m++; m_el[3][2] = *m++;
        m_el[0][3] = *m++; m_el[1][3] = *m++; m_el[2][3] = *m++; m_el[3][3] = *m;
    }

    /**
	 * Set the matrix to these 16 explicit values.
	 */
    void setValue(SimdScalar xx, SimdScalar xy, SimdScalar xz, SimdScalar xw,
                  SimdScalar yx, SimdScalar yy, SimdScalar yz, SimdScalar yw,
                  SimdScalar zx, SimdScalar zy, SimdScalar zz, SimdScalar zw,
                  SimdScalar wx, SimdScalar wy, SimdScalar wz, SimdScalar ww) {
        m_el[0][0] = xx; m_el[0][1] = xy; m_el[0][2] = xz; m_el[0][3] = xw;
        m_el[1][0] = yx; m_el[1][1] = yy; m_el[1][2] = yz; m_el[1][3] = yw;
        m_el[2][0] = zx; m_el[2][1] = zy; m_el[2][2] = zz; m_el[2][3] = zw;
        m_el[3][0] = wx; m_el[3][1] = wy; m_el[3][2] = wz; m_el[3][3] = ww;
    }
	
	/**
	 * Scale the columns of this matrix with x, y, z, w respectively. 
	 */
    void scale(SimdScalar x, SimdScalar y, SimdScalar z, SimdScalar w) {
        m_el[0][0] *= x; m_el[0][1] *= y; m_el[0][2] *= z; m_el[0][3] *= w;
        m_el[1][0] *= x; m_el[1][1] *= y; m_el[1][2] *= z; m_el[1][3] *= w;
        m_el[2][0] *= x; m_el[2][1] *= y; m_el[2][2] *= z; m_el[2][3] *= w;
        m_el[3][0] *= x; m_el[3][1] *= y; m_el[3][2] *= z; m_el[3][3] *= w;
    }

	/**
	 * Return a column-scaled version of this matrix.
	 */
    SimdMatrix4x4 scaled(SimdScalar x, SimdScalar y, SimdScalar z, SimdScalar w) const {
        return SimdMatrix4x4(m_el[0][0] * x, m_el[0][1] * y, m_el[0][2] * z, m_el[0][3] * w,
                            m_el[1][0] * x, m_el[1][1] * y, m_el[1][2] * z, m_el[1][3] * w,
                            m_el[2][0] * x, m_el[2][1] * y, m_el[2][2] * z, m_el[2][3] * w,
                            m_el[3][0] * x, m_el[3][1] * y, m_el[3][2] * z, m_el[3][3] * w);
    }

	/**
	 * Set this matrix to I.
	 */
    void setIdentity() { 
        setValue(SimdScalar(1.0), SimdScalar(0.0), SimdScalar(0.0), SimdScalar(0.0),
                 SimdScalar(0.0), SimdScalar(1.0), SimdScalar(0.0), SimdScalar(0.0),
                 SimdScalar(0.0), SimdScalar(0.0), SimdScalar(1.0), SimdScalar(0.0),
			     SimdScalar(0.0), SimdScalar(0.0), SimdScalar(0.0), SimdScalar(1.0)); 
    }

	/**
	 * Read the element from row i, column j.
	 */
	float getElement(int i, int j) {
		return (float) m_el[i][j];
	}
	
    /**
	 * Copy the contents to a contiguous block of 16 floats.
	 */
    void getValue(float *m) const {
        *m++ = (float) m_el[0][0]; *m++ = (float) m_el[1][0]; *m++ = (float) m_el[2][0]; *m++ = (float) m_el[3][0];
        *m++ = (float) m_el[0][1]; *m++ = (float) m_el[1][1]; *m++ = (float) m_el[2][1]; *m++ = (float) m_el[3][1];
        *m++ = (float) m_el[0][2]; *m++ = (float) m_el[1][2]; *m++ = (float) m_el[2][2]; *m++ = (float) m_el[3][2];
        *m++ = (float) m_el[0][3]; *m++ = (float) m_el[1][3]; *m++ = (float) m_el[2][3]; *m = (float) m_el[3][3];
    }

    /**
	 * Copy the contents to a contiguous block of 16 doubles.
	 */
    void getValue(double *m) const {
        *m++ = m_el[0][0]; *m++ = m_el[1][0]; *m++ = m_el[2][0]; *m++ = m_el[3][0];
        *m++ = m_el[0][1]; *m++ = m_el[1][1]; *m++ = m_el[2][1]; *m++ = m_el[3][1];
        *m++ = m_el[0][2]; *m++ = m_el[1][2]; *m++ = m_el[2][2]; *m++ = m_el[3][2];
        *m++ = m_el[0][3]; *m++ = m_el[1][3]; *m++ = m_el[2][3]; *m = m_el[3][3];
    }

	/** 
	 * Left-multiply this matrix with the argument.
	 */
    SimdMatrix4x4& operator*=(const SimdMatrix4x4& m); 

	/**
	 * Left-multiply column c with row vector c.
	 */
    SimdScalar tdot(int c, const SimdVector4& v) const {
        return m_el[0][c] * v[0]
			+ m_el[1][c] * v[1]
			+ m_el[2][c] * v[2]
			+ m_el[3][c] * v[3];
    }

	/* I'll postpone this for now... - nzc*/ 
/*      SimdScalar    determinant() const; */
/*  	SimdMatrix4x4 adjoint() const; */
/*      SimdMatrix4x4 inverse() const;  */

	SimdMatrix4x4 absolute() const;

	SimdMatrix4x4 transposed() const; 
	void         transpose();

	void         invert();
  
  
	/** Orthogonal projection matrix constructor. */
   static SimdMatrix4x4 orthogonalProjection(
        double            left,
        double            right,
        double            bottom,
        double            top,
        double            nearval,
        double            farval);

	/** Perspective projection matrix constructor. */
    static SimdMatrix4x4 perspectiveProjection(
        double            left,
        double            right,
        double            bottom,
        double            top,
        double            nearval,
        double            farval);
	
	
protected:
	/**
	 * Access with [row index][column index]
	 */
    SimdVector4 m_el[4];
};

/* These multiplicators do exactly what you ask from them: they
 * multiply in the indicated order. */
SimdVector4   operator*(const SimdMatrix4x4& m, const SimdVector4& v);
SimdVector4   operator*(const SimdVector4& v, const SimdMatrix4x4& m);
SimdMatrix4x4 operator*(const SimdMatrix4x4& m1, const SimdMatrix4x4& m2);


/*
 * This is a supposedly faster inverter than the cofactor
 * computation. It uses an LU decomposition sort of thing.  */
SIMD_FORCE_INLINE void SimdMatrix4x4::invert()  {
	/* normalize row 0 */

	int i,j,k;

	for (i=1; i < 4; i++) m_el[0][i] /= m_el[0][0];
	for (i=1; i < 4; i++)  { 
		for (j=i; j < 4; j++)  { // do a column of L
			SimdScalar sum = 0.0;
			for (k = 0; k < i; k++)  
				sum += m_el[j][k] * m_el[k][i];
			m_el[j][i] -= sum;
		}
		if (i == 3) continue;
		for (j=i+1; j < 4; j++)  {  // do a row of U
			SimdScalar sum = 0.0;
			for (k = 0; k < i; k++)
				sum += m_el[i][k]*m_el[k][j];
			m_el[i][j] = 
				(m_el[i][j]-sum) / m_el[i][i];
		}
	}
	for (i = 0; i < 4; i++ )  // invert L
		for (j = i; j < 4; j++ )  {
			SimdScalar x = 1.0;
			if ( i != j ) {
				x = 0.0;
				for (k = i; k < j; k++ ) 
					x -= m_el[j][k]*m_el[k][i];
			}
			m_el[j][i] = x / m_el[j][j];
		}
	for (i = 0; i < 4; i++ )   // invert U
		for (j = i; j < 4; j++ )  {
			if ( i == j ) continue;
			SimdScalar sum = 0.0;
			for (k = i; k < j; k++ )
				sum += m_el[k][j]*( (i==k) ? 1.0 : m_el[i][k] );
			m_el[i][j] = -sum;
		}
	for (i = 0; i < 4; i++ )   // final inversion
		for (j = 0; j < 4; j++ )  {
			SimdScalar sum = 0.0;
			for (k = ((i>j)?i:j); k < 4; k++ )  
				sum += ((j==k)?1.0:m_el[j][k])*m_el[k][i];
			m_el[j][i] = sum;
		}
}

/* We do things slightly different here, because the invert() modifies
 * the buffer itself. This makes it impossible to make this op right
 * away. Like other, still missing facilities, I will repair this
 * later. */
/* SIMD_FORCE_INLINE T_Matrix4x4 SimdMatrix4x4::inverse() const */
/*  { */
/*  } */


SIMD_FORCE_INLINE SimdMatrix4x4& SimdMatrix4x4::operator*=(const SimdMatrix4x4& m)
{
	setValue(m.tdot(0, m_el[0]), m.tdot(1, m_el[0]), m.tdot(2, m_el[0]), m.tdot(3, m_el[0]),
             m.tdot(0, m_el[1]), m.tdot(1, m_el[1]), m.tdot(2, m_el[1]), m.tdot(3, m_el[1]),
             m.tdot(0, m_el[2]), m.tdot(1, m_el[2]), m.tdot(2, m_el[2]), m.tdot(3, m_el[2]),
             m.tdot(0, m_el[3]), m.tdot(1, m_el[3]), m.tdot(2, m_el[3]), m.tdot(3, m_el[3]));
    return *this;

}

SIMD_FORCE_INLINE SimdVector4 operator*(const SimdMatrix4x4& m, const SimdVector4& v) {
    return SimdVector4(dot(m[0], v), dot(m[1], v), dot(m[2], v), dot(m[3], v));
}

SIMD_FORCE_INLINE SimdVector4 operator*(const SimdVector4& v, const SimdMatrix4x4& m) {
    return SimdVector4(m.tdot(0, v), m.tdot(1, v), m.tdot(2, v), m.tdot(3, v));
}

SIMD_FORCE_INLINE SimdMatrix4x4 operator*(const SimdMatrix4x4& m1, const SimdMatrix4x4& m2) {
	return 
		SimdMatrix4x4(m2.tdot(0, m1[0]), m2.tdot(1, m1[0]), m2.tdot(2, m1[0]), m2.tdot(3, m1[0]),
                     m2.tdot(0, m1[1]), m2.tdot(1, m1[1]), m2.tdot(2, m1[1]), m2.tdot(3, m1[1]),
                     m2.tdot(0, m1[2]), m2.tdot(1, m1[2]), m2.tdot(2, m1[2]), m2.tdot(3, m1[2]),
                     m2.tdot(0, m1[3]), m2.tdot(1, m1[3]), m2.tdot(2, m1[3]), m2.tdot(3, m1[3]));
}


SIMD_FORCE_INLINE SimdMatrix4x4 SimdMatrix4x4::transposed() const {
    return SimdMatrix4x4(m_el[0][0], m_el[1][0], m_el[2][0], m_el[3][0],
                        m_el[0][1], m_el[1][1], m_el[2][1], m_el[3][1],
                        m_el[0][2], m_el[1][2], m_el[2][2], m_el[3][2],
                        m_el[0][3], m_el[1][3], m_el[2][3], m_el[3][3]);
}

SIMD_FORCE_INLINE void SimdMatrix4x4::transpose() {
	*this = transposed();
}

SIMD_FORCE_INLINE SimdMatrix4x4 SimdMatrix4x4::absolute() const {
    return 
        SimdMatrix4x4(fabs(m_el[0][0]), fabs(m_el[0][1]), fabs(m_el[0][2]), fabs(m_el[0][3]),
                     fabs(m_el[1][0]), fabs(m_el[1][1]), fabs(m_el[1][2]), fabs(m_el[1][3]),
                     fabs(m_el[2][0]), fabs(m_el[2][1]), fabs(m_el[2][2]), fabs(m_el[2][3]),
                     fabs(m_el[3][0]), fabs(m_el[3][1]), fabs(m_el[3][2]), fabs(m_el[3][3]));
}

SIMD_FORCE_INLINE SimdMatrix4x4 SimdMatrix4x4::orthogonalProjection(
    double            left,
    double            right,
    double            bottom,
    double            top,
    double            nearval,
    double            farval) {

    // Adapted from Mesa
    double x, y, z;
    double tx,  ty, tz;

    x = 2.0 / (right-left);
    y = 2.0 / (top-bottom);
    z = -2.0 / (farval-nearval);
    tx = -(right+left) / (right-left);
    ty = -(top+bottom) / (top-bottom);
    tz = -(farval+nearval) / (farval-nearval);

    return 
        SimdMatrix4x4( x , 0.0, 0.0,  tx,
                	 0.0,  y , 0.0,  ty,
                	 0.0, 0.0,  z ,  tz,
                	 0.0, 0.0, 0.0, 1.0);
}


SIMD_FORCE_INLINE SimdMatrix4x4 SimdMatrix4x4::perspectiveProjection(
    double left,    
    double right,
    double bottom,  
    double top,
    double nearval, 
    double farval) {

    double x, y, a, b, c, d;

    x = (2.0*nearval) / (right-left);
    y = (2.0*nearval) / (top-bottom);
    a = (right+left) / (right-left);
    b = (top+bottom) / (top-bottom);

    if (farval >= SIMD_INFINITY) {
       // Infinite view frustum
       c = -1.0;
       d = -2.0 * nearval;
    } else {
       c = -(farval+nearval) / (farval-nearval);
       d = -(2.0*farval*nearval) / (farval-nearval);
    }

    return SimdMatrix4x4(
        x,  0,  a,  0,
        0,  y,  b,  0,
        0,  0,  c,  d,
        0,  0, -1,  0);
}

#endif

