

#include "Module.h"

#include "Pipeline.h"



using namespace grn;


Module::Module( Pipeline* aParent, int aX, int aY )
	: iX(aX)
	, iY(aY)
	, iSizeX(60)
	, iSizeY(60)
	, rPipeline(aParent)
{
}


Module::~Module() 
{
	// Pad destruction
	for ( std::vector<OutputPad*>::iterator it = iOutputs.begin()
		; it != iOutputs.end()
		; it++ )
		{
			delete *it;
		}
	iOutputs.clear();
		
	for ( std::vector<InputPad*>::iterator it = iInputs.begin()
		; it != iInputs.end()
		; it++ )
		{
			rPipeline->RemoveLinksTo(*it);		
			delete *it;
		}
	iInputs.clear();
}


void Module::Select()
{
	assert (rPipeline);
	
	rPipeline->SetSelected(this);
}


void Module::addPad( InputPad* aPad )
{
	iInputs.push_back(aPad);
}


void Module::addPad( OutputPad* aPad )
{
	iOutputs.push_back(aPad);
}


void Module::RemovePad( InputPad* aPad )
{
	InputPadList::iterator it = find(iInputs.begin(),iInputs.end(), aPad);
	assert( it!=iInputs.end() );
	
	iInputs.erase(it);
	
	rPipeline->RemoveLinksTo(aPad);		
	
	delete aPad;
}


void Module::RemovePad( OutputPad* aPad )
{
	OutputPadList::iterator it = find(iOutputs.begin(),iOutputs.end(), aPad);
	assert( it!=iOutputs.end() );
	
	iOutputs.erase(it);
	
	delete aPad;
}


void Module::RemoveLinksTo( InputPad* aPad )
{
	for ( std::vector<OutputPad*>::iterator it = iOutputs.begin()
		; it != iOutputs.end()
		; it++ )
		{
			OutputPad* lP = (*it);
			for ( int i=0 ; i < lP->GetNumLinks(); )
			{
				if (lP->GetLink(i)==aPad)
				{
					// move following links one position
					for ( int j=i+1; j<lP->iNumLinks; j++ ) lP->iLinks[j-1] = lP->iLinks[j];
					// decrease number of links
					lP->iNumLinks--;
					// we don shrink the memory, but this is not a problem.
				}
				else i++;
			}
		}
}


Module::OutputPad* Module::FindOutputPad( const std::string& aName )
{
	for ( std::vector<OutputPad*>::iterator it = iOutputs.begin()
		; it != iOutputs.end()
		; it++ )
		if ((*it)->iName.compare(aName)==0) return (*it);
	return NULL;
}


Module::InputPad* Module::FindInputPad( const std::string& aName )
{
	for ( std::vector<InputPad*>::iterator it = iInputs.begin()
		; it != iInputs.end()
		; it++ )
		if ((*it)->iName.compare(aName)==0) return (*it);
	return NULL;
}


void Module::Serialize( std::ostream& s ) const
{
	s << getModuleTypeName() <<std::endl;
	s << iX << " " << iY << std::endl;
}


void Module::Deserialize( std::istream& s )
{
	// name has been read already
	s>>iX; 
	s>>iY;
}


Module::LinkList Module::GetOutgoingLinks() const
{
	Module::LinkList res;
	
	for ( std::vector<OutputPad*>::const_iterator it = iOutputs.begin()
		; it != iOutputs.end()
		; it++ )
		{
			for ( int i=0; i<(*it)->GetNumLinks(); i++ ) 
				res.push_back( std::pair<OutputPad*,InputPad*> (*it,(*it)->GetLink(i)) );
		}
			
	return res;
}


void Module::ClearData()
{
	for ( std::vector<InputPad*>::iterator it = iInputs.begin()
		; it != iInputs.end()
		; it++ )
		(*it)->iDataQueue.clear();
}


bool Module::AllInputsHaveData()
{
	bool lResult = true;
	for ( std::vector<InputPad*>::iterator it = iInputs.begin()
		; it != iInputs.end()
		; it++ )
		lResult &= (*it)->hasData();
		
	return lResult;
}


bool Module::AllInputsHaveEOD()
{
	// Check if all inputs have an end-of-data
	bool lEODs = true;
	for ( std::vector<InputPad*>::iterator it = iInputs.begin()
		; it != iInputs.end()
		; it++ )
		lEODs &= (*it)->hasEOD();
		
	return lEODs;
}


bool Module::Eod()
{	
	bool lConsumed = false;
	
	while (AllInputsHaveEOD())
	{
		// Remove EODs
		for ( std::vector<InputPad*>::iterator it = iInputs.begin()
			; it != iInputs.end()
			; it++ )
			(*it)->getGenericData(true);
		
		// Emit to all outputs
		for ( std::vector<OutputPad*>::iterator it = iOutputs.begin()
			; it != iOutputs.end()
			; it++ )
			rPipeline->EmitEOD(*it);
			
		lConsumed = true;
	}
	
	return lConsumed;
}


bool Module::InputPad::hasData()
{ 
	// We have data and it is not an EOD
	return (iDataQueue.size()>0) && (*iDataQueue.begin());
}


bool Module::InputPad::hasEOD()
{
	// We have data and it is an EOD
	return (iDataQueue.size()>0) && (!(*iDataQueue.begin()));
}


