#ifndef _GRN_MODULE_H_
#define _GRN_MODULE_H_

#include <vector>
#include <string>
#include <cassert>
#include <list>
#include <iostream>

#include "Pointer.h"
#include "DataType.h"
#include "Visitor.h"


#define DECLARE_MODULE(X) 	public:																	\
							static const std::string sName;											\
							typedef tFactory<X> Factory; 											\
							virtual const std::string& getModuleTypeName() const { return sName; }	\
							virtual void Accept( BaseVisitor* aV ) { AcceptImpl( this, aV); }
							
#define IMPLEMENT_MODULE(X) const std::string grn::X::sName(#X);



namespace grn
{

	class Pipeline;
	class Engine;


	class Module
	{
	public:
	
		/** @{ Visitor pattern methods 
		 */

		template < class ModuleType > 
		static void AcceptImpl( ModuleType *aVisited, BaseVisitor *aVisitor )
		{
			Visitor< ModuleType > *lVisitorImpl = dynamic_cast< Visitor< ModuleType >* >( aVisitor );
			if( lVisitorImpl ) lVisitorImpl->Visit( aVisited );
			else 
			{
				// Try if it accepts generic modules
				Visitor<Module> *lVisitorImpl = dynamic_cast< Visitor<Module>* >( aVisitor );
				if( lVisitorImpl ) lVisitorImpl->Visit( aVisited );
			}
		}

		virtual void Accept( BaseVisitor* aV ) { AcceptImpl( this, aV); }

		/** @} */

		
		class Pad
		{
		public: 
			Pad( DataType aType, std::string aName ) 
				: iType(aType)
				, iName(aName)
				{}
			virtual ~Pad() {}	

			std::string iType;
			std::string iName;		
		};

		
		class InputPad : public Pad
		{
		public: 
			InputPad( DataType aType, std::string aName, Module* aOwner ) 
				: Pad(aType, aName)
				, rModule(aOwner)
				{
					strncpy(tName,aName.c_str(),80);
				}

			/** Data queue for a pad. The data can be actual objects or EODs that are represented by NULL pointers. */
			typedef std::list< Pointer<DataObject> > DataQueue;
			DataQueue iDataQueue;
			Module* rModule;
			
			// Queries
			bool hasData();
			bool hasEOD();
			
			Pointer<DataObject> getGenericData( bool aRemove=false ) 
			{ 
				Pointer<DataObject> lRes = 0;
				if (iDataQueue.size())
				{
					lRes = (*iDataQueue.begin());
					if (aRemove) iDataQueue.erase(iDataQueue.begin());
				}
				return lRes;
			}

			// temp buffer for blender UI edition of name
			char tName[80];
		};		
	
		template<class D> class tInputPad : public InputPad
		{
		public:
			tInputPad( std::string aName, Module* rOwner ) : InputPad(D::Type, aName, rOwner ) { rOwner->addPad(this); }

			Pointer<D> getData( bool aRemove=false ) 
			{ 
				return down_cast<D> (getGenericData(aRemove));
			}
		};


		class OutputPad : public Pad
		{
		public: 
			OutputPad( DataType aType, std::string aName ) 
				: Pad(aType, aName)
				, iNumLinks(0)
				, iLinks(0) 
				{}

			void Link( InputPad* aP )
			{
				iNumLinks++;
				iLinks = (InputPad**) realloc( iLinks, sizeof(InputPad*)*iNumLinks );
				iLinks[iNumLinks-1] = aP;
			}
			
			inline short GetNumLinks() const { return iNumLinks; }
			inline InputPad* GetLink( int i ) const { assert((i>=0)&&(i<iNumLinks)); return iLinks[i]; }
			
			/** Weird blender link management. */
			short iNumLinks;
			InputPad** iLinks;
		};		

		template<class D> class tOutputPad : public OutputPad
		{
		public:
			tOutputPad( std::string aName, Module* rOwner ) : OutputPad(D::Type, aName ) { rOwner->addPad(this); }
		};


	public:
		
		class Factory
		{
		public: 
			virtual ~Factory() {}
			virtual Module* New( Pipeline* aParent ) const = 0;
			virtual const std::string& getName() const = 0;
		};

		template<class M> class tFactory : public Factory
		{
		public: 
			virtual M* New( Pipeline* aParent ) const { return new M(aParent); }
			virtual const std::string& getName() const { return M::sName; }
		};
		
		
	public:
	
		Module( Pipeline* aParent, int aX=30, int aY=30 );
		virtual ~Module();
		
		void addPad( InputPad* aPad );
		void addPad( OutputPad* aPad );
		void RemovePad( InputPad* aPad );
		void RemovePad( OutputPad* aPad );

		/** Remove any links to the given input pad from the output pads of this module. */
		void RemoveLinksTo( InputPad* aPad );
		
		void setSize( int aX, int aY ) { iSizeX = aX; iSizeY = aY; }
		virtual const std::string& getModuleTypeName() const = 0;

		/** Remove any data from queues. */
		virtual void ClearData();
		
		/** Return true if there is at least one object in each input pad. */
		bool AllInputsHaveData();
		
		/** Return true if all inputs have an EOD signal. */
		bool AllInputsHaveEOD();
		
		/** Called first with the engine that will run. */
		virtual void Link( Engine* ) {}
		
		/** Called once before any initialization or run. */
		virtual void Setup() {}
		
		/** Called once before a set of runs of the pipeline. */
		virtual void Init() {}
		
		/** Called every time we receive data from another module of the pipeline. */
		virtual void Run() {}
		
		/** Called every time an input pad has received an end-of-data signal. */
		virtual bool Eod();
		
		/** Output serializer.*/
		virtual void Serialize( std::ostream& s ) const;

		/** Input serializer.*/
		virtual void Deserialize( std::istream& s );

		typedef std::vector< std::pair<OutputPad*,InputPad*> > LinkList;

		/** Return the list of output links. 
		 * Warning: This is used to serialize and the order matters. 
		 */
		LinkList GetOutgoingLinks() const;

		/** Get the parent pipeline. */
		Pipeline* GetParent() const { return rPipeline; };
		
		/** Get a user readable name. */
		virtual std::string GetModuleUserName() { return getModuleTypeName(); }
		
	public:
		OutputPad* FindOutputPad( const std::string& aName );
		InputPad* FindInputPad( const std::string& aName );
		
		void Select();
		
	public:
		
		typedef std::vector<InputPad*> InputPadList;
		InputPadList iInputs;
		
		typedef std::vector<OutputPad*> OutputPadList;
		OutputPadList iOutputs;
		
		int iX,iY;
		int iSizeX,iSizeY;

	protected:
		
		Pipeline* rPipeline;

	};

}

#endif /*_MODULE_H_*/
