
#include "Pipeline.h"

#include "ModuleDraw.h"

//#include "engine/Program.h" 			// Check GL stuff

#include <ios>
#include <fstream>
#include <algorithm>

#include "GironaLog.h"


using namespace grn;


IMPLEMENT_MODULE(Pipeline);


Pipeline::Pipeline( Pipeline* aPipeline)
: Module(aPipeline)
{
	iSelected = NULL;
	iNextDataTypeID = 0;
	iSizeX = 180;
	iSizeY = 120;
	iName = "unnamed";
	iStartEnabled = true;
	iEnabled = true;
	iLazyToggleEnabled = false;
}


void Pipeline::Draw( ModuleDrawVisitor& lDrawer )
{
	for ( ModuleList::iterator it = iModules.begin()
		; it != iModules.end()
		; it++ )
		{
			lDrawer.iSelected = ((*it) == iSelected);
			(*it)->Accept( &lDrawer );
		}
}

	
void Pipeline::SetSelected( Module* aModule )
{
	iSelected = aModule;
}

	
void Pipeline::RemoveSelected()
{
	iModules.erase( std::remove(iModules.begin(), iModules.end(), iSelected), iModules.end() );
	delete iSelected;
	iSelected = NULL;
}

	
Pipeline::TypeMap& Pipeline::GetTypeMap()
{
	// Look in parent
	if (rPipeline) return rPipeline->GetTypeMap();
	else return iTypeMap;
}

	
int Pipeline::GetDataTypeID( std::string aType )
{
	// Look in parent
	if (rPipeline) return rPipeline->GetDataTypeID(aType);
	else
	{
		// We are the root
		if ( iTypeMap.find(aType)==iTypeMap.end() ) iTypeMap[aType] = iNextDataTypeID++;
		return iTypeMap[aType];
	}
}


DataType Pipeline::FindDataType( int aID )
{
	// Look in parent
	if (rPipeline) return rPipeline->FindDataType(aID);
	else
	{
		// We are the root
		for ( std::map< DataType, int >::iterator it = iTypeMap.begin()
			; it != iTypeMap.end()
			; it++ )
			if (it->second==aID) return it->first;
			
		return "Unknown";
	}
}


Module::Factory* Pipeline::FindFactory( std::string aModuleType )
{
	// Look in parent
	if (rPipeline) return rPipeline->FindFactory(aModuleType);
	else
	{
		// We are the root
		bool lFound = false;
		for ( ModuleFactoryList::iterator it = iModuleFactories.begin()
			; (it!=iModuleFactories.end()) && (!lFound)
			; it++ )
			if ( (*it)->getName() == aModuleType ) return (*it);
			
		return 0;
	}
}


DataObject::Factory* Pipeline::FindObjectFactory( std::string aDataType )
{
	// Look in parent
	if (rPipeline) return rPipeline->FindObjectFactory(aDataType);
	else
	{
		// We are the root
		bool lFound = false;
		for ( ObjectFactoryList::iterator it = iObjectFactories.begin()
			; (it!=iObjectFactories.end()) && (!lFound)
			; it++ )
			if ( (*it)->getName() == aDataType ) return (*it);
			
		return 0;
	}
}


void Pipeline::Setup()
{
	iEnabled = iStartEnabled;
	
	log("Setting up pipeline");

	for ( ModuleList::iterator it = iModules.begin()
		; it != iModules.end()
		; it++ )
		(*it)->Setup();
		
//	CHECK_GL_ERROR(" [GRN] After setup : ");	    

}


void Pipeline::Run()
{
	if (!iEnabled) return;
//	if (!AllInputsHaveData()) return;
	
	log("Initialising pipeline: ",GetModuleUserName());

//	CHECK_GL_ERROR(" [GRN] Before init : ");	    

	// Init all modules in this pipeline
	// "Import" modules while insert data from external incoming links here
	for ( ModuleList::iterator it = iModules.begin()
		; it != iModules.end()
		; it++ )
	{
		//log("  Initing : ",(*it)->getModuleTypeName());
		(*it)->Init();
	}
//	CHECK_GL_ERROR(" [GRN] After init : ");	    
	
	log("Running pipeline");

	while ( iRunQueue.size() )
	{
		while ( iRunQueue.size() )		
		{
			Module* lC = iRunQueue.front();
			iRunQueue.pop_front();
	
			log("  Running : ",lC->getModuleTypeName());
		
			lC->Run();
			bool lReRun = lC->Eod();
			if (lReRun) iRunQueue.push_back(lC);
		}
	}
	
//	CHECK_GL_ERROR(" [GRN] Before clear : ");	    

	for ( ModuleList::iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
		(*it)->ClearData();
	
	log("End of pipeline");
	
}


void Pipeline::ClearData()
{
	Module::ClearData();

	for ( ModuleList::iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
		(*it)->ClearData();
}


void Pipeline::Emit( Module::OutputPad* aPad, DataObject* aData )
{
	for (int i=0; i<aPad->GetNumLinks(); i++)
	{
		Module::InputPad* lInPad = aPad->GetLink(i);
		lInPad->iDataQueue.push_back(aData);
		if (*iRunQueue.rbegin()!=lInPad->rModule)
			iRunQueue.push_back(lInPad->rModule);
	}
}


void Pipeline::EmitEOD( Module::OutputPad* aPad )
{
	//log4(" EOD emmited by :",aPad->iName," of type: ",aPad->iType);
	
	for (int i=0; i<aPad->GetNumLinks(); i++)
	{
		Module::InputPad* lInPad = aPad->GetLink(i);
		lInPad->iDataQueue.push_back(NULL);
		if (*iRunQueue.rbegin()!=lInPad->rModule)
			iRunQueue.push_back(lInPad->rModule);
	}
}


void Pipeline::Clear()
{
	Module::ClearData();
	
	for ( ModuleList::iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
		delete *it;
	
	iModules.clear();
	iRunQueue.clear();
	iSelected = NULL;
}


void Pipeline::RemoveLinksTo( Module::InputPad* aPad )
{
	for ( ModuleList::iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
		(*it)->RemoveLinksTo(aPad);	
}


void Pipeline::Load( const std::string& aFileName )
{
	Clear();

	std::ifstream s( aFileName.c_str(), std::ios::in );
	if (!s.is_open()) return;

	// Read the first module type name, to skip it.	
	std::string dummy;
	s>>dummy;

	Deserialize( s );	
}


void Pipeline::Deserialize( std::istream& s )
{
	Module::Deserialize(s);
	
	
	std::string id;
	s>>id;

	int version;
	s>>version;
	
	// name
	s>>iName;
	
	if (version>1)  s>>iStartEnabled;
	else iStartEnabled = true;
	
	// Modules
	int nummodules;
	s>>nummodules;
	
	for (int i=0;i<nummodules;i++)
	{
		std::string lModuleName;
		s>>lModuleName;
		
		// Look for module factory
		Module::Factory* lFactory = FindFactory(lModuleName);		
		if (lFactory)
		{
			Module* lM = lFactory->New(this);

			// Read it
			lM->Deserialize(s);
							
			// Add it
			iModules.push_back(lM);
			
		}
		else std::cout <<" [GRN] Load error: Cannot create module : "<<lModuleName<<std::endl;
	}
	
	
	// Links
	int lSourceModule, lDestModule=-1;
	s>>lSourceModule;
	while ((lSourceModule>=0) && (!s.eof()))
	{
		std::string lSourceLink,lDestLink;

		s>>lSourceLink;
		s>>lDestModule;
		s>>lDestLink;
		
		std::cout<< " [GRN] Linking: "<<lSourceLink<<" in "<<lSourceModule<<" with "<<lDestLink<<" in "<<lDestModule<<std::endl;
		
		if ( (lSourceModule<0) || (lSourceModule>=(int)iModules.size())) std::cout<< " [GRN] Load error: Cannot create link. Source module error."<<std::endl;
		else if ( (lDestModule<0) || (lDestModule>=(int)iModules.size())) std::cout<< " [GRN] Load error: Cannot create link. Dest module error."<<std::endl;
		else 
		{
			Module::OutputPad* lOut = iModules[lSourceModule]->FindOutputPad(lSourceLink);
			Module::InputPad* lIn = iModules[lDestModule]->FindInputPad(lDestLink);
			if ((!lOut)||(!lIn)) std::cout<< " [GRN] Load error: Cannot create link. Pad error."<<std::endl;
			else
			{
				// Create the link
				lOut->Link(lIn);
			}
		}

		// next module
		s>>lSourceModule;
	}
}


void Pipeline::Save( const std::string& aFileName ) const
{
	std::ofstream s( aFileName.c_str(), std::ios::out );	
	Serialize(s);
}


void Pipeline::Serialize( std::ostream& s ) const
{
	Module::Serialize(s);
	
	// ID
	s<<"GironaPipeline"<<std::endl;
	// version
	s<<"2"<<std::endl;
	
	// name
	s<<iName<<std::endl;
	
	s<<iStartEnabled<<std::endl;
	
	// number of modules
	s<<iModules.size()<<std::endl;

	// module list
	int i=0;
	std::map<Module*,int> iMapIDs;
	for ( ModuleList::const_iterator it = iModules.begin()
		; it!= iModules.end()
		; it++, i++ )
	{
		iMapIDs[(*it)] = i;
		(*it)->Serialize(s);
		s<<std::endl;
	}
	
	// links
	for ( ModuleList::const_iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
	{
		Module::LinkList l = (*it)->GetOutgoingLinks();
		for ( Module::LinkList::iterator itt = l.begin()
			; itt != l.end()
			; itt++ )
		{
			// Source module and pad
			s<<iMapIDs[*it]<<" "<<itt->first->iName;
			
			// Destination module and pad
			s<<" "<<iMapIDs[itt->second->rModule]<<" "<<itt->second->iName<<std::endl;
		}
	}
	
	// end of links mark
	s<<(int)-1<<std::endl;
}


void Pipeline::Link( Engine* aEngine )
{
	std::cout<<"Linking pipeline"<<std::endl;

	for ( ModuleList::iterator it = iModules.begin()
		; it != iModules.end()
		; it++ )
		(*it)->Link(aEngine);
		
}


Pipeline* Pipeline::FindPipeline( std::string aName )
{
	if (rPipeline) return rPipeline->FindPipeline( aName );
	else return 0;
}




