
// NOTE: Based on somebody else's smart pointer

#ifndef _GRN_POINTER_H_
#define _GRN_POINTER_H_

#include <cassert>

#include <vector>


namespace grn
{


/**
  * \brief Base reference counter
  * Used by all objects that are managed through smart pointers
  * Once reference count reaches zero, deletes itself.
  * <b>You must NEVER delete a reference counted object directly!
  * All access to a reference counted object MUST be done through
  * the smart pointer interface as offered by the Pointer template class</b>
  */
class RefCounter
{
	private:

		/*! Reference count */
		int                                           m_iRefCount;


	public:

		/**
		* Set reference count to 0
		*/
		inline                                        RefCounter();

		/**
		* Dummy method for virtual destructor chain
		*/
		inline virtual                               ~RefCounter();

		/**
		* Increase reference count
		*/
		inline void                                   IncRef();

		/**
		* Decrease reference count and garbage collect if 0
		*/
		inline void                                   DecRef();

		/**
		* \return                                     Current reference count
		*/
		inline int                                    GetRefCount() const;
};


/**
  * \brief Smart pointer template for reference counted classes
  * <b>Reference counted objects MUST be accessed through smart pointers
  * at all times to make sure reference counting scheme operates correctly.
  * Failure to do so will result in strange and unpredictable behaviour and crashes.</b>
  * The smart pointer class takes care of reference counting management at
  * creation, destruction and assignement of pointers, and lets the reference counting
  * scheme correctly free objects that are no longer referenced.
  */
template < class T > class Pointer
{
	private:

		/*! Pointer to object */
		T                                            *m_pkObject;

	public:

		/**
		* Assign pointer and increase reference for object if not null
		* \param pkObject               Pointer to object
		*/
		inline                                        Pointer( T *pkObject = 0 );

		/**
		* Assign pointer and increment reference for object if not null
		* \param rkPointer                            Pointer object ref
		*/
		inline                                        Pointer( const Pointer< T > &rkPointer );

		/**
		* Decrease reference for object if not null
		*/
		inline                                       ~Pointer();

		/**
		* \return                                     true if pointer is valid, false if null
		*/
		inline                                        operator const bool () const;

		/**
		* \return                                     Typecast to pointer to class
		*/
		inline                                        operator T* ();

		/**
		* \return                                     Typecast to pointer to class
		*/
		inline                                        operator const T* () const;

		/**
		* \return                                     Typecast to pointer to class and dereference
		*/
		inline T                                     &operator * ();

		/**
		* \return                                     Typecast to pointer to class and dereference
		*/
		inline const T                               &operator * () const;

		/**
		* \return                                     Typecast to pointer to class and dereference
		*/
		inline T                                     *operator -> ();

		/**
		* \return                                     Typecast to pointer to class and dereference
		*/
		inline const T                               *operator -> () const;

		/**
		* \return                                     True if invalid pointer (null), false if valid
		*/
		inline bool                                   operator !() const;

		/**
		* Assignment of pointer, decrease reference count for old object and increase for current object
		* \param rkPointer                            Smart pointer to copy from
		* \return                                     Pointer ref (this)
		*/
		inline Pointer< T >                          &operator = ( const Pointer< T > &rkPointer );

		/**
		* Assignment of pointer, decrease reference count for old object and increase for current object
		* \param pkObject                             Pointer to object
		* \return                                     Pointer ref (this)
		*/
		inline Pointer< T >                          &operator = ( T *pkObject );

		/**
		* \param rkPointer                            Reference smart pointer object
		* \return                                     True if pointers are equal, false if not
		*/
		inline bool                                   operator == ( const Pointer< T > &rkPointer ) const;

		/**
		* \param pkObject                             Pointer to reference object
		* \return                                     True if pointers are equal, false if not
		*/
		inline bool                                   operator == ( const T *pkObject ) const;

		/**
		* \param rkPointer                            Reference smart pointer object
		* \return                                     True if pointers are not equal, false if equal
		*/
		inline bool                                   operator != ( const Pointer< T > &rkPointer ) const;

		/**
		* \param pkObject                             Pointer to reference object
		* \return                                     True if pointers are not equal, false if equal
		*/
		inline bool                                   operator != ( const T *pkObject ) const;

		/**
		* \param rkPointer                            Reference smart pointer object
		* \return                                     True if pointer address is less that reference , false if equal
		*/
		inline bool                                   operator < ( const Pointer< T > &rkPointer ) const;

		/**
		* \param pkObject                             Pointer to reference object
		* \return                                     True if pointer address is less that reference, false if equal
		*/
		inline bool                                   operator < ( const T *pkObject ) const;
};




inline RefCounter::RefCounter() :
	m_iRefCount( 0 )
{
}


inline RefCounter::~RefCounter()
{
}


inline void RefCounter::IncRef()
{
	++m_iRefCount;
}


inline void RefCounter::DecRef()
{
	if( !--m_iRefCount )
		delete this;
}


inline int RefCounter::GetRefCount() const
{
	return m_iRefCount;
}





template < class T > inline Pointer< T >::Pointer( T *pkObject ) :
	m_pkObject( pkObject )
{
	if( m_pkObject )
		m_pkObject->IncRef();
}


template < class T > inline Pointer< T >::Pointer( const Pointer< T > &rkPointer ) :
	m_pkObject( rkPointer.m_pkObject )
{
	if( m_pkObject )
		m_pkObject->IncRef();
}


template < class T > inline Pointer< T >::~Pointer()
{
	if( m_pkObject )
		m_pkObject->DecRef(), m_pkObject = 0;
}


template < class T > inline Pointer< T >::operator const bool () const
{
	return( m_pkObject != 0 );
}


template < class T > inline Pointer< T >::operator T* ()
{
	return m_pkObject;
}


template < class T > inline Pointer< T >::operator const T* () const
{
	return m_pkObject;
}


template < class T > inline T &Pointer< T >::operator * ()
{
	assert( m_pkObject );
	return *m_pkObject;
}


template < class T > inline const T &Pointer< T >::operator * () const
{
	assert( m_pkObject );
	return *m_pkObject;
}

template < class T > inline T *Pointer< T >::operator -> ()
{
	return m_pkObject;
}


template < class T > inline const T *Pointer< T >::operator -> () const
{
	return m_pkObject;
}


template < class T > inline bool Pointer< T >::operator !() const
{
	return( m_pkObject == 0 );
}


template < class T > inline Pointer< T > &Pointer< T >::operator = ( const Pointer< T > &rkPointer )
{
	T *pkOldObject = m_pkObject;

	m_pkObject = rkPointer.m_pkObject;

	if( m_pkObject  )
		m_pkObject->IncRef();
	
	if( pkOldObject )
		pkOldObject->DecRef();

	return( *this );
}


template < class T > inline Pointer< T > &Pointer< T >::operator = ( T *pkObject )
{
	T *pkOldObject = m_pkObject;

	m_pkObject = pkObject;

	if( m_pkObject )
		m_pkObject->IncRef();

	if( pkOldObject )
		pkOldObject->DecRef();

	return( *this );
}


template < class T > inline bool Pointer< T >::operator == ( const T *pkObject ) const
{
	return( m_pkObject == pkObject );
}


template < class T > inline bool Pointer< T >::operator != ( const T *pkObject ) const
{
	return( m_pkObject != pkObject );
}


template < class T > inline bool Pointer< T >::operator < ( const T *pkObject ) const
{
	return( m_pkObject < pkObject );
}


template < class T > inline bool Pointer< T >::operator == ( const Pointer< T > &rkPointer ) const
{
	return( m_pkObject == rkPointer.m_pkObject );
}


template < class T > inline bool Pointer< T >::operator != ( const Pointer< T > &rkPointer ) const
{
	return( m_pkObject != rkPointer.m_pkObject );
}


template < class T > inline bool Pointer< T >::operator < ( const Pointer< T > &rkPointer ) const
{
	return( m_pkObject < rkPointer.m_pkObject );
}


};


#endif
