
#include "RootPipeline.h"
#include "Serialise.h"

#include "datatypes/BasicTypes.h"
#include "datatypes/Object.h"
#include "datatypes/Mesh.h"
#include "datatypes/Material.h"
#include "datatypes/Camera.h"
#include "datatypes/SpotLight.h"
#include "datatypes/PhysicsEnvironment.h"

//////////////////////////////////////////////
///////// GENERIC
//////////////////////////////////////////////

#include "modules/generic/Import.h"
#include "modules/generic/Export.h"
#include "modules/generic/MarkGroup.h"
#include "modules/generic/MakeGroup.h"
#include "modules/generic/UnGroup.h"
#include "modules/generic/SequenceSelect.h"
#include "modules/generic/If.h"
#include "modules/generic/Replicate.h"
#include "modules/generic/EndOfData.h"

//////////////////////////////////////////////
///////// PHYSICS
//////////////////////////////////////////////

#include "modules/physics/AddToPhysicsEnv.h"
#include "modules/physics/CreatePhysicsEnv.h"
#include "modules/physics/UpdatePhysicsEnv.h"
#include "modules/physics/CollisionQuery.h"

//////////////////////////////////////////////
///////// LOGIC
//////////////////////////////////////////////
#include "modules/logic/TimeStep.h"
#include "modules/logic/KeyEvent.h"
#include "modules/logic/SelectObject.h"
#include "modules/logic/SetObjectPosition.h"
#include "modules/logic/PhysicsModifier.h"
#include "modules/logic/ConstantVector.h"

#include "modules/logic/VariableVector.h"
#include "modules/logic/AddVector.h"

#include "modules/logic/ConstantInteger.h"
#include "modules/logic/VariableInteger.h"
#include "modules/logic/AddInteger.h"

#include "modules/logic/ConstantReal.h"
#include "modules/logic/VariableReal.h"
#include "modules/logic/AddReal.h"
#include "modules/logic/CompareReal.h"

#include "modules/logic/MultiplyVector.h"
#include "modules/logic/ClampVector.h"
#include "modules/logic/EnablePipeline.h"

#include "modules/logic/Not.h"
#include "modules/logic/ConstantBoolean.h"
#include "modules/logic/VariableBoolean.h"
#include "modules/logic/OperateBoolean.h"

//////////////////////////////////////////////
///////// GRAPHICS
//////////////////////////////////////////////

#include "modules/graphics/Fade.h"
#include "modules/graphics/DepthMaterial.h"
#include "modules/graphics/UVMaterial.h"
#include "modules/graphics/SetCameraPlanes.h"
#include "modules/graphics/PlanarMirror.h"
#include "modules/graphics/OrbitCamera.h"
#include "modules/graphics/SpotCamera.h"
#include "modules/graphics/SetMaterialTexture.h"
#include "modules/graphics/SetMaterialCamera.h"
#include "modules/graphics/Screen.h"
#include "modules/graphics/OffscreenTarget.h"
#include "modules/graphics/MeshRender.h"
#include "modules/graphics/PresentRenderTarget.h"
#include "modules/graphics/TextureBlit.h"
#include "modules/graphics/GetRenderTexture.h"


//////////////////////////////////////////////
///////// SCENE
//////////////////////////////////////////////

#include "modules/scene/ObjectParts.h"
#include "modules/scene/GetBoundingVolume.h"
#include "modules/scene/ObjectForm.h"
#include "modules/scene/LinkHierarchy.h"



#include "GironaLog.h"

#include <ios>
#include <fstream>
#include <algorithm>


using namespace grn;


IMPLEMENT_MODULE(RootPipeline);


RootPipeline::RootPipeline()
: Pipeline( 0 )
{

	// Register data types
	GRN_RegisterDataType( Boolean );
	GRN_RegisterDataType( Integer );
	GRN_RegisterDataType( Real );
	GRN_RegisterDataType( Vector );
	GRN_RegisterDataType( Group );
	GRN_RegisterDataType( String );
	GRN_RegisterDataType( StringList );
	
	GRN_RegisterDataType( Transform );
	GRN_RegisterDataType( BoundingVolume );
	GRN_RegisterDataType( Object );
	GRN_RegisterDataType( Mesh );
	GRN_RegisterDataType( Camera );
	GRN_RegisterDataType( SpotLight );
	GRN_RegisterDataType( Material );
	GRN_RegisterDataType( PhongMaterial );
	GRN_RegisterDataType( RenderTarget );
	GRN_RegisterDataType( ScreenRenderTarget );
	GRN_RegisterDataType( OffscreenRenderTarget );
	GRN_RegisterDataType( Texture );
	
	GRN_RegisterDataType( PhysicsEnvironment );
	GRN_RegisterDataType( PhysicsProperties );
	GRN_RegisterDataType( ScenePhysicsProperties );
	GRN_RegisterDataType( CollisionList );
	
	// Register modules
	iModuleFactories.push_back( new Pipeline::Factory() );
	iModuleFactories.push_back( new Import::Factory() );
	iModuleFactories.push_back( new Export::Factory() );
	iModuleFactories.push_back( new MarkGroup::Factory() );
	iModuleFactories.push_back( new MakeGroup::Factory() );
	iModuleFactories.push_back( new UnGroup::Factory() );
	iModuleFactories.push_back( new SequenceSelect::Factory() );
	iModuleFactories.push_back( new If::Factory() );
	iModuleFactories.push_back( new Replicate::Factory() );
	iModuleFactories.push_back( new EndOfData::Factory() );
	
	iModuleFactories.push_back( new CreatePhysicsEnv::Factory() );
	iModuleFactories.push_back( new AddToPhysicsEnv::Factory() );
	iModuleFactories.push_back( new UpdatePhysicsEnv::Factory() );
	iModuleFactories.push_back( new CollisionQuery::Factory() );

	iModuleFactories.push_back( new TimeStep::Factory() );
	iModuleFactories.push_back( new KeyEvent::Factory() );
	iModuleFactories.push_back( new SelectObject::Factory() );
	iModuleFactories.push_back( new SetObjectPosition::Factory() );
	iModuleFactories.push_back( new PhysicsModifier::Factory() );
	
	iModuleFactories.push_back( new ConstantVector::Factory() );
	iModuleFactories.push_back( new VariableVector::Factory() );
	iModuleFactories.push_back( new AddVector::Factory() );
	
	iModuleFactories.push_back( new ConstantInteger::Factory() );
	iModuleFactories.push_back( new VariableInteger::Factory() );
	iModuleFactories.push_back( new AddInteger::Factory() );
	
	iModuleFactories.push_back( new ConstantReal::Factory() );
	iModuleFactories.push_back( new VariableReal::Factory() );
	iModuleFactories.push_back( new AddReal::Factory() );
	iModuleFactories.push_back( new CompareReal::Factory() );

	iModuleFactories.push_back( new ConstantBoolean::Factory() );
	iModuleFactories.push_back( new VariableBoolean::Factory() );
	iModuleFactories.push_back( new OperateBoolean::Factory() );
	
	iModuleFactories.push_back( new MultiplyVector::Factory() );
	iModuleFactories.push_back( new ClampVector::Factory() );
	iModuleFactories.push_back( new EnablePipeline::Factory() );
	iModuleFactories.push_back( new Not::Factory() );
 
	iModuleFactories.push_back( new ObjectParts::Factory() );
	iModuleFactories.push_back( new GetBoundingVolume::Factory() );
	iModuleFactories.push_back( new ObjectForm::Factory() );
	iModuleFactories.push_back( new LinkHierarchy::Factory() );
	
	iModuleFactories.push_back( new Fade::Factory() );
	iModuleFactories.push_back( new DepthMaterial::Factory() );
	iModuleFactories.push_back( new UVMaterial::Factory() );
	iModuleFactories.push_back( new Screen::Factory() );
	iModuleFactories.push_back( new OffscreenTarget::Factory() );
	iModuleFactories.push_back( new MeshRender::Factory() );
	iModuleFactories.push_back( new PresentRenderTarget::Factory() );
	iModuleFactories.push_back( new TextureBlit::Factory() );
	iModuleFactories.push_back( new PlanarMirror::Factory() );
	iModuleFactories.push_back( new OrbitCamera::Factory() );
	iModuleFactories.push_back( new SpotCamera::Factory() );
	iModuleFactories.push_back( new SetMaterialTexture::Factory() );
	iModuleFactories.push_back( new SetMaterialCamera::Factory() );
	iModuleFactories.push_back( new SetCameraPlanes::Factory() );
	iModuleFactories.push_back( new GetRenderTexture::Factory() );

}



void RootPipeline::SetExecutionStage( Pipeline* aPipeline, ExecutionStage aStage, int aOrder )
{
	if (!aPipeline) return;
	
	// Remove from previous group
	iBuildPipelines.remove( aPipeline );
	iRenderPipelines.remove( aPipeline );
		
	bool lFound = false;
	for ( std::multimap<int, Pipeline*>::iterator it = iLogicPipelines.begin()
		; (it != iLogicPipelines.end()) && (!lFound)
		; )
		if (it->second==aPipeline)
		{
			iLogicPipelines.erase(it);
			lFound=true;
			//log2(" -------------- Removing with order = ",it->first);
		}
		else it++;
	
	//log2(" -------------- SetExecutionOrder = ",aOrder);

	// Add to new group
	switch (aStage)
	{
		case eBuild: iBuildPipelines.push_back(aPipeline); break;
		case eLogic: iLogicPipelines.insert( std::pair<int,Pipeline*>(aOrder,aPipeline) ); break;
		case eRender: iRenderPipelines.push_back(aPipeline); break;
		default: break;
	}
}


RootPipeline::ExecutionStage RootPipeline::GetExecutionStage( Pipeline* aPipeline ) const
{
	
	for ( std::multimap<int, Pipeline*>::const_iterator it = iLogicPipelines.begin()
		; it != iLogicPipelines.end()
		; it++)
		if (it->second==aPipeline) return eLogic;
	
	
	if ( std::find(iBuildPipelines.begin(), iBuildPipelines.end(), aPipeline) != iBuildPipelines.end() ) return eBuild;
	else if ( std::find(iRenderPipelines.begin(), iRenderPipelines.end(), aPipeline) != iRenderPipelines.end() ) return eRender;
	else return eNone;
}


int RootPipeline::GetExecutionOrder( Pipeline* aPipeline ) const 
{
	int lRes = 0;
	for ( std::multimap<int, Pipeline*>::const_iterator it = iLogicPipelines.begin()
		; it != iLogicPipelines.end()
		; it++)
		if (it->second==aPipeline) lRes = it->first;

	//log2(" -------------- GetExecutionOrder = ",lRes);

	return lRes;
}


void RootPipeline::Serialize( std::ostream& s ) const
{
	Pipeline::Serialize(s);
	
	s << iBuildPipelines.size() << std::endl;
	for ( std::list<Pipeline*>::const_iterator i=iBuildPipelines.begin(); i!=iBuildPipelines.end(); i++ ) s << (*i)->GetName() << std::endl;

	s << iLogicPipelines.size() << std::endl;
	
	for ( std::multimap<int,Pipeline*>::const_iterator i=iLogicPipelines.begin(); i!=iLogicPipelines.end(); i++ ) 
	{
		s << i->first << std::endl;
		s << i->second->GetName() << std::endl;
	}

	s << iRenderPipelines.size() << std::endl;
	for ( std::list<Pipeline*>::const_iterator i=iRenderPipelines.begin(); i!=iRenderPipelines.end(); i++ ) s << (*i)->GetName() << std::endl;
	
}


void RootPipeline::Deserialize( std::istream& s )
{
	Pipeline::Deserialize(s);
	
	int n = -1;
	std::string name;

	s >> n;
	for (int i=0;i<n;i++) 
	{
		s >> name;
		iBuildPipelines.push_back( FindPipeline( name ) );
	}
	
	s >> n;
	for (int i=0;i<n;i++) 
	{
		int lOrder = 0;
		s >> lOrder;
		s >> name;
		iLogicPipelines.insert( std::pair<int,Pipeline*>(lOrder,FindPipeline( name )) );
	}
	
	s >> n;
	for (int i=0;i<n;i++)  
	{
		s >> name;
		iRenderPipelines.push_back( FindPipeline( name ) );
	}
	
}


Pipeline* RootPipeline::FindPipeline( std::string aName )
{
	for ( unsigned int i=0; i<iModules.size(); i++ )
	{
		Pipeline* lP = dynamic_cast<Pipeline*>(iModules[i]);
		if (lP)
		{
			if (lP->GetName()==aName) return lP;
		}
	}
	
	std::cerr<<" Didn find : "<<aName<<std::endl;
	assert( false && "Pipeline not found by name");
	
	return 0;
}


void RootPipeline::SaveRuntime( const std::string& aFileName ) const
{
	std::ofstream s( aFileName.c_str(), std::ios::out );	

	Module::Serialize(s);
	
	// ID
	s<<"GironaPipeline"<<std::endl;
	// version
	s<<"2"<<std::endl;
	
	// name
	s<<iName<<std::endl;
	
	s<<iStartEnabled<<std::endl;
	
	// number of modules
	s<<iModules.size()-iBuildPipelines.size()<<std::endl;

	// module list
	int i=0;
	std::map<Module*,int> lMapIDs;
	for ( ModuleList::const_iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
	{
		// Discard build pipelines
		if ( GetExecutionStage(static_cast<Pipeline*>(*it)) != eBuild )
		{			
			lMapIDs[(*it)] = i;
			(*it)->Serialize(s);
			s<<std::endl;
			i++;
		}
		
	}
	
	// links
	for ( ModuleList::const_iterator it = iModules.begin()
		; it!= iModules.end()
		; it++ )
	{
		Module::LinkList l = (*it)->GetOutgoingLinks();
		for ( Module::LinkList::iterator itt = l.begin()
			; itt != l.end()
			; itt++ )
		{
			// Source module and pad
			s<<lMapIDs[*it]<<" "<<itt->first->iName;
			
			// Destination module and pad
			s<<" "<<lMapIDs[itt->second->rModule]<<" "<<itt->second->iName<<std::endl;
		}
	}
	
	// end of links mark
	s<<(int)-1<<std::endl;
	
	// Root pipeline info
	s << 0 << std::endl;

	s << iLogicPipelines.size() << std::endl;	
	for ( std::multimap<int,Pipeline*>::const_iterator i=iLogicPipelines.begin(); i!=iLogicPipelines.end(); i++ ) 
	{
		s << i->first << std::endl;
		s << i->second->GetName() << std::endl;
	}

	s << iRenderPipelines.size() << std::endl;
	for ( std::list<Pipeline*>::const_iterator i=iRenderPipelines.begin(); i!=iRenderPipelines.end(); i++ ) s << (*i)->GetName() << std::endl;
	
}


