
#include "Serialise.h"

#include "GironaLog.h"
#include "engine/Renderer.h"

using namespace grn;


SerialiseVisitor::SerialiseVisitor( const std::string& aFileName )
 : iStream(aFileName)
{	
	// Write header
	const char* lHeader = "Girona Data File\n";

	char lDataTypeName[64];
	bzero( lDataTypeName, 64*sizeof(char) );
	strncpy( lDataTypeName, lHeader, 64 );
	iStream.Write( lDataTypeName, 64*sizeof(char) );

	const int lVersion = 1;
	iStream<<lVersion;
}


void SerialiseVisitor::Flush()
{
	// Flush references
	
	// Iterate because adding one of the missing objects can add more references to the dictionary
	while (iDictionary.size() != iDone.size() )
	{
		bool lFound = false;
		for ( Dictionary::iterator it=iDictionary.begin()
			; (it!=iDictionary.end()) && (!lFound)
			; it++ )
			if ( find( iDone.begin(), iDone.end(), it->second ) == iDone.end() )
			{
				lFound = true;
				it->first->Accept(this);
			}
	}
}


SerialiseVisitor::~SerialiseVisitor()
{
}


SerialiseVisitor::BinStream::BinStream( const std::string& aFileName )
{
	iFile = fopen(aFileName.c_str(),"wb");
}


SerialiseVisitor::BinStream::~BinStream()
{
	if (iFile)
	{
		fclose(iFile);
		iFile = NULL;
	}
}


int SerialiseVisitor::FindOrAdd(DataObject* aO)
{
	// NULL object? NULL ID!
	if (!aO) return -1;
	
	std::map<DataObject*,int>::iterator it = iDictionary.find(aO);
	if ( it != iDictionary.end() ) return it->second;
	else
	{
		int id = iDictionary.size()+1;
		iDictionary[aO] = id;
		return id;
	}

}


void SerialiseVisitor::SerialiseInputs( Pipeline* aPipeline )
{
	//For each input pad
	for ( Module::InputPadList::iterator it2 = aPipeline->iInputs.begin()
		; it2 != aPipeline->iInputs.end()
		; it2++ )
	{

		// Write number of objects
		int lNumObjs = (*it2)->iDataQueue.size();
		
		iStream<<lNumObjs;
		
		// Write each object using the visitor pattern call
		for ( Module::InputPad::DataQueue::iterator it3 = (*it2)->iDataQueue.begin()
			; it3 != (*it2)->iDataQueue.end()
			; it3++ )
			WriteReference((*it3));
	}
		
}


void SerialiseVisitor::SerialiseHeader( DataObject* aO )
{
	assert (aO);
	
	char lDataTypeName[64];
	bzero(lDataTypeName,64*sizeof(char));
	strncpy(lDataTypeName,aO->GetDataType().c_str(),64);
	
	iStream.Write(lDataTypeName,64*sizeof(char));
	
	int lID = FindOrAdd( aO );
	
	assert( std::find( iDone.begin(), iDone.end(), lID ) == iDone.end() );
	iDone.push_back(lID);
	
	iStream<<lID;

	log( std::string("    [serialise] ")+lDataTypeName+std::string(" with ID "), lID);
	
}


void SerialiseVisitor::Visit( Boolean* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iValue;
}


void SerialiseVisitor::Visit( Integer* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iValue;
}


void SerialiseVisitor::Visit( Real* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iValue;
}


void SerialiseVisitor::Visit( Vector* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iValue;
}


void SerialiseVisitor::Visit( BoundingVolume* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iAABB;
}


void SerialiseVisitor::Visit( Color* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iColor;
}


void SerialiseVisitor::Visit( Transform* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iTransform.getOrigin();
	iStream << aO->iTransform.getRotation();
}


void SerialiseVisitor::Visit( String* aO )
{
	SerialiseHeader( aO );
	iStream << aO->iValue;
}


void SerialiseVisitor::Visit( StringList* aO )
{
	SerialiseHeader( aO );

	iStream << aO->iValue.size();
	for (unsigned int i=0;i<aO->iValue.size();i++)
		iStream << aO->iValue[i];
}


void SerialiseVisitor::Visit( Group* aO )
{
	SerialiseHeader( aO );

	iStream << aO->iCollections.size();
	for (unsigned int i=0;i<aO->iCollections.size();i++)
	{
		iStream << aO->iCollections[i].iName;
		iStream << aO->iCollections[i].iObjects.size();
		for (unsigned int j=0; j<aO->iCollections[i].iObjects.size(); j++)
			WriteReference(aO->iCollections[i].iObjects[j]);
	}
}


void SerialiseVisitor::Visit( Camera* aO )
{
	SerialiseHeader( aO );

	iStream << aO->iProjection;
	iStream << aO->iPosition;
	iStream << aO->iTarget;
	iStream << aO->iUp;
	iStream << aO->iNear;
	iStream << aO->iFar;
	iStream << aO->iTop;
	iStream << aO->iBottom;
	iStream << aO->iLeft;
	iStream << aO->iRight;
	iStream << aO->iFov;
}


void SerialiseVisitor::Visit( PhongMaterial* aO )
{
	SerialiseHeader( aO );

	iStream << aO->iExponent;
	iStream << aO->iSpecular;
	iStream << aO->iDiffuse;
	iStream << aO->iAmbient;

}


void SerialiseVisitor::Visit( Mesh* aO )
{
	SerialiseHeader( aO );

	iStream << aO->iNumFaces;
	iStream << aO->iNumVertices;

	iStream << aO->iIndexBuffer->iDataType;
	iStream << aO->iIndexBuffer->iSize;
	iStream.Write( aO->iIndexBuffer->iData, aO->iIndexBuffer->iSize );

	iStream << aO->iVertexBuffers.size();
	for ( unsigned int i=0; i<aO->iVertexBuffers.size(); i++ )
	{
		iStream << aO->iVertexBuffers[i]->iDataType;
		iStream << aO->iVertexBuffers[i]->iNumElements;
		iStream << aO->iVertexBuffers[i]->iSemantic;
		iStream << aO->iVertexBuffers[i]->iSize;
		iStream.Write( aO->iVertexBuffers[i]->iData, aO->iVertexBuffers[i]->iSize );
	}

	iStream << aO->iAABB;

}


void SerialiseVisitor::Visit( SpotLight* aO )
{
	SerialiseHeader( aO );

	WriteReference(aO->iTransform);
	iStream << aO->iColor;
	iStream << aO->iAngle;
}


void SerialiseVisitor::Visit( Object* aO )
{
	SerialiseHeader( aO );
	
	WriteReference(aO->iName);
	WriteReference(aO->iMesh);
	WriteReference(aO->iMaterial);
	WriteReference(aO->iParent);
	WriteReference(aO->iParentName);
	WriteReference(aO->iProperties);
	WriteReference(aO->iInstanceTransform);
	iStream << aO->iAABB;;
	
}


void SerialiseVisitor::Visit( PhysicsEnvironment* aO )
{
	SerialiseHeader( aO );
	
	iStream<<aO->iRawData.size();
	
	for ( unsigned int i=0; i<aO->iRawData.size(); i++ )
	{
		WriteReference( aO->iRawData[i].first );
		WriteReference( aO->iRawData[i].second );
	}
		
}


void SerialiseVisitor::Visit( PhysicsProperties* aO )
{
	SerialiseHeader( aO );
	
	iStream<<aO->iActor;
	iStream<<aO->iDynamic;
	iStream<<aO->iRigidBody;
	iStream<<aO->iGhost;
	iStream<<aO->iDeformable;

	iStream<<aO->iMass;
	iStream<<aO->iFriction;
	iStream<<aO->iRestitution;
	iStream<<aO->iInertia;
	iStream<<aO->iRadius;
	iStream<<aO->iLinearDrag;
	iStream<<aO->iAngularDrag;
	iStream<<aO->iHasAnisotropicFriction;
	iStream<<aO->iAnisotropicFrictionScaling;
	iStream<<aO->iDoFH;
	iStream<<aO->iDoRotFH;
		
	iStream<<aO->iLocalBoundsCenter;
	iStream<<aO->iLocalBoundsExtends;
		
	iStream<<aO->iBoundType;
	iStream<<aO->iConcave;
}


/////////////////////////////////////////////////
/////////////////////////////////////////////////
/////////////////////////////////////////////////


Deserialiser::Deserialiser( const std::string& aFileName, Pipeline* aPipeline, Engine* aEngine )
 : iStream(aFileName)
 , rPipeline(aPipeline)
 , rEngine(aEngine)
{	
	// Read file header
	char lDataTypeName[64];
	bzero( lDataTypeName, 64*sizeof(char) );
	iStream.Read( lDataTypeName, 64*sizeof(char) );

	log( "    [deserialise] header : ",lDataTypeName);

	int lVersion = 0;
	iStream>>lVersion;
	assert (lVersion==1);
}

void Deserialiser::Flush()
{	
	log( "    [deserialise] flush ");

	// Read all data objects
	while (!iStream.eof())
	{
		// Read object header
		char lDataTypeName[64];
		iStream.Read(lDataTypeName,64*sizeof(char));
		int lID = -1;
		iStream>>lID;

		log( std::string("    [deserialise] ")+std::string(lDataTypeName)+" with ID ", lID);
		assert(iDictionary.find( lID ) == iDictionary.end() );
		
		// Create the suitable type
		Pointer<DataObject> lAO = rPipeline->FindObjectFactory(lDataTypeName)->New();
		
		// Read internal data
		lAO->Accept(this);
		
		// Register the pointer with the ID
		iDictionary[lID] = lAO;		
	}
	
	
	log( "    [deserialise] link ");

	// Link references
	Linker lLinker(*this);
	
	for ( Dictionary::iterator it = iDictionary.begin()
		; it != iDictionary.end()
		; it++ )
		it->second->Accept( &lLinker);
			

	log( "    [deserialise] emit ");

	// Send to module inputs
	for ( InputObjects::iterator it = iInputObjects.begin()
		; it != iInputObjects.end()
		; it++ )
	{
		for ( std::vector<int>::iterator it2 = it->second.begin()
			; it2 != it->second.end()
			; it2++ )
		{
			assert ( iDictionary.find(*it2) != iDictionary.end() );
			it->first->iDataQueue.push_back( iDictionary[*it2] );
		}
	}
	
}


void Deserialiser::DeserialiseInputs( Pipeline* aPipeline )
{
	//For each input pad
	for ( Module::InputPadList::iterator it2 = aPipeline->iInputs.begin()
		; it2 != aPipeline->iInputs.end()
		; it2++ )
	{

		// Read number of objects
		int lNumObjs = 0;
		iStream>>lNumObjs;
		
		iInputObjects[*it2] = std::vector<int>();
		for ( int i=0; i<lNumObjs; i++ )
		{	
			int lID = -1;
			iStream>>lID;
			iInputObjects[*it2].push_back(lID);	
		}
	}	
}


Deserialiser::BinStream::BinStream( const std::string& aFileName )
{
	iFile = fopen(aFileName.c_str(),"rb");
}


Deserialiser::BinStream::~BinStream()
{
	if (iFile)
	{
		fclose(iFile);
		iFile = NULL;
	}
}


void Deserialiser::Visit( Boolean* aO )
{
	iStream >> aO->iValue;
}


void Deserialiser::Visit( Integer* aO )
{
	iStream >> aO->iValue;
}


void Deserialiser::Visit( Real* aO )
{
	iStream >> aO->iValue;
}


void Deserialiser::Visit( Vector* aO )
{
	iStream >> aO->iValue;
}


void Deserialiser::Visit( BoundingVolume* aO )
{
	iStream >> aO->iAABB;
}


void Deserialiser::Visit( Color* aO )
{
	iStream >> aO->iColor;
}


void Deserialiser::Visit( Transform* aO )
{
	SimdPoint3 p;
	SimdQuaternion r;
	iStream >> p;
	iStream >> r;
	aO->iTransform = SimdTransform(r,p);
}


void Deserialiser::Visit( String* aO )
{
	iStream >> aO->iValue;
}


void Deserialiser::Visit( StringList* aO )
{
	int size;
	iStream >> size;
	for (int i=0;i<size;i++)
	{
		std::string t;
		iStream >> t;
		aO->iValue.push_back(t);
	}
}


void Deserialiser::Visit( Group* aO )
{
	int size;
	iStream >> size;
	for (int i=0;i<size;i++)
	{
		std::string lName;
		iStream >> lName;
		aO->iCollections.push_back(Group::Collection(lName));

		int iObjectSize;
		iStream >> iObjectSize;
		for (int j=0; j<iObjectSize; j++)
		{
			aO->iCollections[i].iObjects.push_back( Pointer<DataObject>(0) );
			ReadReference(aO);
		}
	}
}


void Deserialiser::Linker::Visit( Group* aO )
{
	for (unsigned int i=0;i<aO->iCollections.size();i++)
		for (unsigned int j=0; j<aO->iCollections[i].iObjects.size(); j++)
		{
			aO->iCollections[i].iObjects[j] = GetReference(aO);
			log("    [relink reference] to ", (int)(DataObject*)aO->iCollections[i].iObjects[j]);
		}
}


void Deserialiser::Visit( Camera* aO )
{
	iStream >> aO->iProjection;
	iStream >> aO->iPosition;
	iStream >> aO->iTarget;
	iStream >> aO->iUp;
	iStream >> aO->iNear;
	iStream >> aO->iFar;
	iStream >> aO->iTop;
	iStream >> aO->iBottom;
	iStream >> aO->iLeft;
	iStream >> aO->iRight;
	iStream >> aO->iFov;
	
	aO->Update();
}


void Deserialiser::Visit( PhongMaterial* aO )
{
	iStream >> aO->iExponent;
	iStream >> aO->iSpecular;
	iStream >> aO->iDiffuse;
	iStream >> aO->iAmbient;

}


void Deserialiser::Visit( Mesh* aO )
{
	iStream >> aO->iNumFaces;
	iStream >> aO->iNumVertices;

	aO->iIndexBuffer = new Mesh::IndexBuffer();
	iStream >> aO->iIndexBuffer->iDataType;
	iStream >> aO->iIndexBuffer->iSize;
	aO->iIndexBuffer->iData = new char[aO->iIndexBuffer->iSize];
	iStream.Read( aO->iIndexBuffer->iData, aO->iIndexBuffer->iSize );

	int iNumBuffers;
	iStream >> iNumBuffers;
	for ( int i=0; i<iNumBuffers; i++ )
	{
		aO->iVertexBuffers.push_back( new Mesh::VertexBuffer() );
		iStream >> aO->iVertexBuffers[i]->iDataType;
		iStream >> aO->iVertexBuffers[i]->iNumElements;
		iStream >> aO->iVertexBuffers[i]->iSemantic;
		iStream >> aO->iVertexBuffers[i]->iSize;
		aO->iVertexBuffers[i]->iData = new char[aO->iVertexBuffers[i]->iSize];
		iStream.Read( aO->iVertexBuffers[i]->iData, aO->iVertexBuffers[i]->iSize );
	}

	iStream >> aO->iAABB;

	rEngine->GetRenderer()->Upload(aO, false);

}


void Deserialiser::Visit( SpotLight* aO )
{
	ReadReference(aO);
	iStream >> aO->iColor;
	iStream >> aO->iAngle;
}

void Deserialiser::Linker::Visit( SpotLight* aO )
{
	aO->iTransform = down_cast<Transform>(GetReference(aO));
}


void Deserialiser::Visit( Object* aO )
{
	ReadReference(aO); // name
	ReadReference(aO); // mesh
	ReadReference(aO); // material
	ReadReference(aO); // parent
	ReadReference(aO); // parent name
	ReadReference(aO); // properties
	ReadReference(aO); // instance transform
	iStream >> aO->iAABB;;	
}

void Deserialiser::Linker::Visit( Object* aO )
{
	aO->iName = down_cast<String>(GetReference(aO));
	aO->iMesh = down_cast<Mesh>(GetReference(aO));
	aO->iMaterial = down_cast<Material>(GetReference(aO));
	aO->iParent = down_cast<Object>(GetReference(aO));
	aO->iParentName = down_cast<String>(GetReference(aO));
	aO->iProperties = down_cast<StringList>(GetReference(aO));
	aO->iInstanceTransform = down_cast<Transform>(GetReference(aO));
	
	if (aO->iParent) aO->iParent->iChildren.push_back(aO);
	
}


void Deserialiser::Visit( PhysicsEnvironment* aO )
{
	int lNum;
	iStream >> lNum;
	
	for ( int i=0; i<lNum; i++ )
	{
		aO->iRawData.push_back( std::pair< Pointer<Object>, Pointer<PhysicsProperties> >(0,0) );
		ReadReference(aO);
		ReadReference(aO);
	}
}


void Deserialiser::Linker::Visit( PhysicsEnvironment* aO )
{
	for ( unsigned int i=0; i<aO->iRawData.size(); i++ )
	{
		aO->iRawData[i].first = down_cast<Object> ( GetReference(aO) );
		aO->iRawData[i].second = down_cast<PhysicsProperties> ( GetReference(aO) );
	}

	aO->Build();	
}


void Deserialiser::Visit( PhysicsProperties* aO )
{
	iStream>>aO->iActor;
	iStream>>aO->iDynamic;
	iStream>>aO->iRigidBody;
	iStream>>aO->iGhost;
	iStream>>aO->iDeformable;

	iStream>>aO->iMass;
	iStream>>aO->iFriction;
	iStream>>aO->iRestitution;
	iStream>>aO->iInertia;
	iStream>>aO->iRadius;
	iStream>>aO->iLinearDrag;
	iStream>>aO->iAngularDrag;
	iStream>>aO->iHasAnisotropicFriction;
	iStream>>aO->iAnisotropicFrictionScaling;
	iStream>>aO->iDoFH;
	iStream>>aO->iDoRotFH;
		
	iStream>>aO->iLocalBoundsCenter;
	iStream>>aO->iLocalBoundsExtends;
		
	iStream>>aO->iBoundType;
	iStream>>aO->iConcave;
}


