

#ifndef GRN_SERIALISE_H_
#define GRN_SERIALISE_H_

#include <map>
#include <cstdio>

#include "datatypes/BasicTypes.h"
#include "datatypes/Camera.h"
#include "datatypes/Material.h"
#include "datatypes/Mesh.h"
#include "datatypes/Object.h"
#include "datatypes/SpotLight.h"
#include "datatypes/PhysicsEnvironment.h"

#include "Pipeline.h"
#include "GironaLog.h"

namespace grn 
{

	class SerialiseVisitor 
		: public Visitor<Boolean>
		, public Visitor<Integer>
		, public Visitor<Real>
		, public Visitor<Color>
		, public Visitor<Vector>
		, public Visitor<Transform>
		, public Visitor<String>
		, public Visitor<StringList>
		, public Visitor<BoundingVolume>
		, public Visitor<Group>
		
		, public Visitor<Camera>
		, public Visitor<PhongMaterial>
		, public Visitor<Mesh>
		, public Visitor<Object>
		, public Visitor<SpotLight>
		
		, public Visitor<PhysicsEnvironment>
		, public Visitor<PhysicsProperties>
		
	{
	public:
	
		SerialiseVisitor( const std::string& aFileName );
		virtual ~SerialiseVisitor();

		void Flush();
	
		void SerialiseInputs( Pipeline* aPipeline );
	
		virtual void Visit( Boolean* aO );
		virtual void Visit( Integer* aO );
		virtual void Visit( Real* aO );
		virtual void Visit( Vector* aO );
		virtual void Visit( Color* aO );
		virtual void Visit( Transform* aO );
		virtual void Visit( String* aO );
		virtual void Visit( StringList* aO );
		virtual void Visit( Group* aO );
		virtual void Visit( BoundingVolume* aO );

		virtual void Visit( Camera* aO );
		virtual void Visit( PhongMaterial* aO );
		virtual void Visit( Mesh* aO );
		virtual void Visit( Object* aO );
		virtual void Visit( SpotLight* aO );
		
		virtual void Visit( PhysicsEnvironment* aO );
		virtual void Visit( PhysicsProperties* aO );
		

	protected:
	
		// Store the IDs of the objects to cross reference in the files.
		typedef std::map<DataObject*,int> Dictionary;
		Dictionary iDictionary;

		// Store the IDs of the objects already written.
		std::vector<int> iDone;
		
		int FindOrAdd( DataObject* aO );
		void SerialiseHeader( DataObject* aO );
		void WriteReference( DataObject* aO ) { int lID = FindOrAdd( aO ); iStream<<lID; }


		class BinStream
		{
		public:
			BinStream( const std::string& aFileName );
			virtual ~BinStream();
			
			template <class T>
			void operator<<( T a )
			{
				assert(iFile);
				fwrite( &a, sizeof(T), 1, iFile );
			}
			
			void operator<<( std::string a )
			{
				assert(iFile);
				(*this) << a.size();
				Write( a.c_str(), sizeof(char)*a.size() );
			}
			
			void Write(const void* aData, int size) { fwrite( aData, size, 1, iFile ); }
			
		protected:
			
			FILE* iFile;
			
		};
	
		BinStream iStream;	
	};



	class Deserialiser
		: public Visitor<Boolean>
		, public Visitor<Integer>
		, public Visitor<Real>
		, public Visitor<Color>
		, public Visitor<Vector>
		, public Visitor<Transform>
		, public Visitor<String>
		, public Visitor<StringList>
		, public Visitor<BoundingVolume>
		, public Visitor<Group>
		
		, public Visitor<Camera>
		, public Visitor<PhongMaterial>
		, public Visitor<Mesh>
		, public Visitor<Object>
		, public Visitor<SpotLight>
		
		, public Visitor<PhysicsEnvironment>
		, public Visitor<PhysicsProperties>

	{
	public:
		Deserialiser( const std::string& aFileName, Pipeline* aPipeline, Engine* aEngine );
		virtual ~Deserialiser() {};

		void Flush();	
		void DeserialiseInputs( Pipeline* aPipeline );

		virtual void Visit( Boolean* aO );
		virtual void Visit( Integer* aO );
		virtual void Visit( Real* aO );
		virtual void Visit( Vector* aO );
		virtual void Visit( Color* aO );
		virtual void Visit( Transform* aO );
		virtual void Visit( String* aO );
		virtual void Visit( StringList* aO );
		virtual void Visit( Group* aO );
		virtual void Visit( BoundingVolume* aO );

		virtual void Visit( Camera* aO );
		virtual void Visit( PhongMaterial* aO );
		virtual void Visit( Mesh* aO );
		virtual void Visit( Object* aO );
		virtual void Visit( SpotLight* aO );
		
		virtual void Visit( PhysicsEnvironment* aO );
		virtual void Visit( PhysicsProperties* aO );

	protected:

		typedef std::map<int,Pointer<DataObject> > Dictionary;
		Dictionary iDictionary;
		
		typedef std::map< DataObject*, std::list<int> > ReferenceList;
		ReferenceList iPendingReferences;
		
		typedef std::map< Module::InputPad*, std::vector<int> > InputObjects;
		InputObjects iInputObjects;
		
		void ReadReference( DataObject* aO ) 
		{ 
			int lID = -1; 
			iStream>>lID; 			
			iPendingReferences[aO].push_back( lID );
		} 

	protected:	
	
		class BinStream
		{
		public:
			BinStream( const std::string& aFileName );
			virtual ~BinStream();
			
			template <class T>
			void operator>>( T& a )
			{
				assert(iFile);
				fread( &a, sizeof(T), 1, iFile );
			}
			
			void operator>>( std::string& a )
			{
				assert(iFile);
				char temp[4096];
				int size;
				
				(*this) >> size;
				assert(size<4095);

				Read( temp, sizeof(char)*size );
				temp[size] = 0;
				a = temp;
			}
			
			void Read(void* aData, int size) { fread( aData, size, 1, iFile ); }
			
			bool eof()
			{
				return feof(iFile);
			}
			
		protected:
			
			FILE* iFile;
			
		};
	
		BinStream iStream;	
		
		Pipeline* rPipeline;
		Engine* rEngine;
		
	protected:
	
		class Linker
			: public Visitor<Group>
			, public Visitor<Object>
			, public Visitor<SpotLight>
			, public Visitor<PhysicsEnvironment>
		{
		public:
			Linker( Deserialiser& aDeserialiser ) : rS(aDeserialiser) {}

			virtual void Visit( Group* aO );
			virtual void Visit( Object* aO );
			virtual void Visit( SpotLight* aO );
			virtual void Visit( PhysicsEnvironment* aO );
			
		protected:
			Deserialiser& rS;
		
			DataObject* GetReference( DataObject* aO )
			{
				assert( rS.iPendingReferences.find(aO) != rS.iPendingReferences.end() );
				
				int lID = rS.iPendingReferences[aO].front();
				rS.iPendingReferences[aO].pop_front();
				
				if (lID<0) return NULL;

				assert( rS.iDictionary.find(lID) != rS.iDictionary.end() );
				return rS.iDictionary[lID];				
			}

		}; // End class linker
			
	}; // End class deserialiser
	
}


#endif 

