
#ifndef CUSCUS_AABB_H
#define CUSCUS_AABB_H

#include <climits>	
#include <map>			// std::pair

#include "SimdPoint3.h"
#include "SimdVector3.h"
#include "SimdMinMax.h"

namespace grn
{
	
	class AABB
	{
		
	public:
	
		AABB() 
		{
			iMin.setX( FLT_MAX );
			iMin.setY( FLT_MAX );
			iMin.setZ( FLT_MAX );
			iMax.setX( -FLT_MAX );
			iMax.setY( -FLT_MAX );
			iMax.setZ( -FLT_MAX );
		};
		
		
		inline void Bound( SimdVector3 aV )
		{
			iMin.setX( SimdMin( iMin.x(), aV.x() ) );
			iMin.setY( SimdMin( iMin.y(), aV.y() ) );
			iMin.setZ( SimdMin( iMin.z(), aV.z() ) );
			iMax.setX( SimdMax( iMax.x(), aV.x() ) );
			iMax.setY( SimdMax( iMax.y(), aV.y() ) );
			iMax.setZ( SimdMax( iMax.z(), aV.z() ) );
		}
		
		inline void Bound( AABB aB )
		{
			Bound(aB.iMin);
			Bound(aB.iMax);
		}
		
		inline float GetLongestBoundedLength() const
		{
			return (iMin.x() == FLT_MAX) ? 1.0f : (iMin-iMax).length();
		}
	
		inline SimdVector3 GetCenter() const
		{
			return (iMin.x() == FLT_MAX) ? SimdVector3(0.0f,0.0f,0.0f) : (iMin+iMax)*0.5f ;
		}
		
		
		SimdVector3 GetSize() const
		{
			return iMax-iMin;
		}


		inline bool ContainsPoint( SimdVector3 p ) const
		{
			return	(p.x()>iMin.x()) && (p.y()>iMin.y()) && (p.z()>iMin.z()) && (p.x()<iMax.x()) && (p.y()<iMax.y()) && (p.z()<iMax.z());
		}
		
		int GetLongestAxis() const
		{
			float lx = iMax.x()-iMin.x();
			float ly = iMax.y()-iMin.y();
			float lz = iMax.z()-iMin.z();
			if ((lx>ly) && (lx>lz)) return 0;
			else if ((ly>lx) && (ly>lz)) return 1;
			else return 2;
		}

			
		void GetCorners( SimdPoint3* aBuf) const
		{
			aBuf[0] = SimdPoint3( iMin.x(), iMin.y(), iMin.z() );
			aBuf[1] = SimdPoint3( iMax.x(), iMin.y(), iMin.z() );
			aBuf[2] = SimdPoint3( iMin.x(), iMax.y(), iMin.z() );
			aBuf[3] = SimdPoint3( iMax.x(), iMax.y(), iMin.z() );
			aBuf[4] = SimdPoint3( iMin.x(), iMin.y(), iMax.z() );
			aBuf[5] = SimdPoint3( iMax.x(), iMin.y(), iMax.z() );
			aBuf[6] = SimdPoint3( iMin.x(), iMax.y(), iMax.z() );
			aBuf[7] = SimdPoint3( iMax.x(), iMax.y(), iMax.z() );
		}

			
		std::pair<AABB,AABB> Subdivide()
		{
			AABB lA = *this;
			AABB lB = *this;
	
			int lAxis = GetLongestAxis();
				
			if (lAxis==0)
			{
				float l = (iMax.x()+iMin.x())/2;
				lA.iMax.setX( l );
				lB.iMin.setX( l );
			}
			else if (lAxis==1)
			{
				float l = (iMax.y()+iMin.y())/2;
				lA.iMax.setY( l );
				lB.iMin.setY( l );
			}
			else
			{
				float l = (iMax.z()+iMin.z())/2;
				lA.iMax.setZ( l );
				lB.iMin.setZ( l );
			}
			
			return std::pair<AABB,AABB> (lA,lB);		
		}
		
		
	public:
		
		SimdVector3 iMin,iMax;	
		
	};


	inline SimdVector3 GetRandomSpherePoint() 
	{
	  SimdVector3 p;
	  float  alfa,beta;
	   
	  alfa=1.0-2.0*((double) rand()/RAND_MAX);  /* real between -1 and 1 */
	  alfa=acos(alfa);                          /* angle between 0 and PI */
	  beta=2*M_PI*((double) rand()/RAND_MAX);   /* angle between 0 and 2*PI */
	   
	  p.setX( sin(alfa)*cos(beta) );      //    p.x()=c.x()+r*sin(alfa)*cos(beta);
	  p.setY( sin(alfa)*sin(beta) );      //    p.y()()=c.y()()+r*sin(alfa)*sin(beta);
	  p.setZ( cos(alfa) );                //    p.z()=c.z()+r*cos(alfa);
	    
	  return p;
	}

} // namespace cus


#endif
