

#include "Camera.h"

#include "SimdTransform.h"
#include "SimdMinMax.h"

#include "GironaLog.h"


#define FOV (45.0f*SIMD_PI/180.0f)

#ifdef GIRONA_PSP
	#define ASPECT (480.0f/272.0f)
#else
	#define ASPECT (1.0f)
#endif

using namespace grn;


const DataType Camera::Type("Girona::Camera");


Camera::Camera( bool iPerspective )
{
	iPosition = SimdVector3( 0.0f, 0.0f, -5.0f );
	iTarget = SimdVector3( 0.0f, 0.0f, 0.0f );
	iUp = SimdVector3( 0.0f, 1.0f, 0.0f );
	iFov = FOV;
	
	if (iPerspective)
	{
		iProjection = ePerspective;
		
		iNear = 1.0f;
		iFar = 100.0f;
	
		ComputeFrustumWindow();
	}
	else
	{
		iProjection = eParallel;
		
		iNear = -0.5f;
		iFar = 0.5f;
		iTop = 0.5f;
		iBottom = -0.5f;
		iRight = 0.5f;
		iLeft = -0.5f;
	}
	
	UpdateMatrices();
}


Camera::~Camera()
{
}


void Camera::ComputeFrustumWindow()
{
	iTop = tan( iFov * 0.5f ) * iNear;
	iBottom = -iTop;
	
	iLeft = ASPECT * iBottom;
	iRight = ASPECT * iTop;
}


void Camera::UpdateMatrices()
{
	
	SimdMatrix4x4 lRot;
	SimdVector3 f = GetDirection();
	SimdVector3 s = f.cross(iUp);
	SimdVector3 u = s.cross(f);
	lRot[0][0] = s.x();
	lRot[0][1] = s.y();
	lRot[0][2]= s.z();
	lRot[0][3] = 0.0;
	
	lRot[1][0] = u.x();
	lRot[1][1] = u.y();
	lRot[1][2] = u.z();
	lRot[1][3] = 0.0;
	
	lRot[2][0] = -f.x();
	lRot[2][1] = -f.y();
	lRot[2][2] = -f.z();
	lRot[2][3] = 0.0;
	
	lRot[3][0] = 0.0;
	lRot[3][1] = 0.0;
	lRot[3][2] = 0.0;
	lRot[3][3] = 1.0;
	
	SimdTransform lT;
	lT.setIdentity();
	lT.setOrigin( SimdPoint3(-iPosition.x(), -iPosition.y(), -iPosition.z()) );
	SimdMatrix4x4 lTrans(lT);

	iWorldToCamera = lRot * lTrans;

	UpdateProjectionMatrices();
}


void Camera::UpdateProjectionMatrices()
{
	if (iProjection==ePerspective) iCameraToClip = SimdMatrix4x4::perspectiveProjection( iLeft, iRight, iBottom, iTop, iNear, iFar );
	else iCameraToClip = SimdMatrix4x4::orthogonalProjection( iLeft, iRight, iBottom, iTop, iNear, iFar );
		
	iClipToWorld = iCameraToClip * iWorldToCamera;
	iClipToWorld.invert();
}


void Camera::ViewVolume( const AABB& aVolume )
{
	// Get current direction
	SimdVector3 lDirection = GetDirection();
	iTarget = SimdVector3(aVolume.GetCenter().x(), aVolume.GetCenter().y(), aVolume.GetCenter().z() );
	
	float lLongestLength = aVolume.GetLongestBoundedLength();
	iPosition = iTarget - lDirection * lLongestLength * 1.5f;
	
	iNear = 1.0f * lLongestLength;
	iFar = 2.0 * lLongestLength;
	
	if (iProjection==ePerspective)
	{
		ComputeFrustumWindow();
	}
	else
	{
		iLeft = -lLongestLength/2.0f;
		iRight = lLongestLength/2.0f;
		iBottom = -lLongestLength/2.0f;
		iTop = lLongestLength/2.0f;
	}
		
	UpdateMatrices();
}


void Camera::Orbit( float aHorizontal, float aVertical )
{
	
	float lAzimuth = aHorizontal * 2.0 * SIMD_PI;
	float lPolar   = aVertical * SIMD_PI;
  
	SimdVector3 kSideDirection = (iTarget-iPosition).cross(iUp);
	kSideDirection.normalize();	
	SimdMatrix3x3 kVerticalRotation = SimdMatrix3x3( SimdQuaternion( kSideDirection, -lPolar ) );
	SimdMatrix3x3 kHorizontalRotation = SimdMatrix3x3( SimdQuaternion( iUp, -lAzimuth ) );
	iPosition = ( (kHorizontalRotation * (kVerticalRotation * (iPosition - iTarget) )) + iTarget  );
	iUp = kVerticalRotation * iUp;
	
	UpdateMatrices();
}


void Camera::Zoom( float aDeltaFactor )
{
	iPosition += (iTarget-iPosition) * aDeltaFactor;

	// Update this to keep seeing the same space	
	iNear = SimdMax( 0.001f, (float) ( iNear - (iTarget-iPosition).length() * aDeltaFactor) );
	iFar -= (iTarget-iPosition).length() * aDeltaFactor;
	
	ComputeFrustumWindow();
	
	UpdateMatrices();
}


SimdVector3 Camera::ClipToWorld( SimdVector3 aClipSpacePoint )
{
	SimdVector4 t = ( iClipToWorld * SimdVector4(aClipSpacePoint.x(), aClipSpacePoint.y(), aClipSpacePoint.z(), 1.0f) );
	return SimdVector3( t.x(), t.y(), t.z() );
}


void Camera::SetDirection( const SimdVector3& aDir )
{
	SimdVector3 lDir = aDir.normalized();
	
	float lDist = (iTarget-iPosition).length();
	
	iPosition = iTarget - (lDir * lDist);
	
	MakeUpVector();
	
	UpdateMatrices();
}


void Camera::SetFromTransform( const SimdTransform& aTrans )
{
	iPosition = aTrans*SimdPoint3(0.0,0.0,0.0);
	iTarget = aTrans*SimdPoint3(0.0,0.0,-1.0);
	iUp = (aTrans*SimdPoint3(0.0,1.0,0.0)) - iPosition;
	UpdateMatrices();
}


void Camera::MakeUpVector()
{
	iUp = iUp.cross(GetDirection()).cross(GetDirection());
	if (iUp.length()<1e-5) iUp = SimdVector3(1.0f,0.0f,0.0f).cross(GetDirection());
	if (iUp.length()<1e-5) iUp = SimdVector3(0.0f,0.0f,1.0f).cross(GetDirection());
}


void Camera::SetPlanesFor( const AABB& aVol )
{
	float lLongestLength = aVol.GetLongestBoundedLength();
	SimdVector3 lCenter = aVol.GetCenter();
	float lDistance = (lCenter-iPosition).length();

	iFar = lDistance+lLongestLength/2;	
	if (lDistance<lLongestLength/2)
	{
		// We are inside
		iNear = lLongestLength * 1.0E-3;		
	}
	else
	{	
		// We are outside
		iNear = lDistance-lLongestLength/2;
	}

	ComputeFrustumWindow();
	UpdateMatrices();
	
}

void Camera::SetFOV( float aFOV )
{
	iFov = aFOV;
	ComputeFrustumWindow();
	UpdateMatrices();
}

