
#ifndef CUSCUS_CAMERA_H
#define CUSCUS_CAMERA_H

#include "../DataType.h"

#include "SimdVector3.h"
#include "SimdMatrix4x4.h"
#include "SimdTransform.h"

#include "AABB.h"


namespace grn
{
	 
	class Camera : public DataObject
	{
		DECLARE_DATATYPE(Camera)

	public:
	
		Camera( bool iPerspective = true );
		virtual ~Camera();
		
		void ViewVolume( const AABB& aVolume );
		
		void Orbit( float aHorizontal, float aVertical );
		
		/** 1.0 leaves the same. */
		void Zoom( float aFactor );
		
		typedef enum { ePerspective, eParallel } Projection;
	
		/** Set the camera parameters according to the given transformation. */
		void SetFromTransform( const SimdTransform& aTrans );
		
		/** Set the camera fov in radians. */
		void SetFOV( float aFOV );
		
		/** Returns position in world space. */
		const SimdVector3& GetPosition() const { return iPosition; }
		
		/** Return the viewing direction in world space. */
		const SimdVector3 GetDirection() const { return (iTarget-iPosition).normalized(); }
		
		/**  Set the viewing direction in world space. The target and distance remains the same, and a new up vector is set. */
		void SetDirection( const SimdVector3& aDir );

		/** Return the up vector in world space. */
		const SimdVector3& GetUpVector() const { return iUp; }
		
		/** Space conversions. */
		SimdVector3 ClipToWorld( SimdVector3 aClipSpacePoint );
		
		/** Return the widtth of the window. World-space distance.*/
		float GetWindowWidth() const { return iRight-iLeft; }
		
		/** Return the concatenation of the view and the projection matrices. */
		inline SimdMatrix4x4 GetViewProjectionMatrix() const { return iCameraToClip * iWorldToCamera; }
		
		inline const SimdMatrix4x4& GetViewMatrix() const { return iWorldToCamera; }
		
		inline const SimdMatrix4x4& GetProjectionMatrix() const { return iCameraToClip; }
		
		/** Adjust near and far planes for the given volume. */
		void SetPlanesFor( const AABB& aVol );
		
		
		Projection iProjection;
		SimdVector3 iPosition, iTarget, iUp;
		float iNear, iFar, iTop, iBottom, iLeft, iRight;
		float iFov;
		
		/** Force recalculation of internal data. */
		void Update() { ComputeFrustumWindow(); UpdateMatrices(); }
		
	protected:

		/** Precomputed matrices stored for efficiency. 
		 * @{
		 */
		SimdMatrix4x4 iWorldToCamera;
		SimdMatrix4x4 iCameraToClip;

		SimdMatrix4x4 iClipToWorld;
		/** @} */
		
	protected:
		void MakeUpVector();
		void UpdateMatrices();
		void UpdateProjectionMatrices();
		void ComputeFrustumWindow();
	};

} // namespace cus

#endif // CAMERA_H
