
#include "engine/Engine.h"
#include "RootPipeline.h"
#include "Serialise.h"

#include "datatypes/BasicTypes.h"
#include "datatypes/Object.h"
#include "datatypes/Mesh.h"
#include "datatypes/Material.h"
#include "datatypes/Camera.h"
#include "datatypes/SpotLight.h"
#include "datatypes/PhysicsEnvironment.h"

#include "GironaLog.h"

#include <ios>
#include <fstream>
#include <algorithm>


using namespace grn;



const char* Engine::sKeyNames[ Engine::_eNumKeys ] = 
			{ "Space", "Left Arrow", "RightArrow", "Up Arrow", "Down Arrow", "eEscape"
			, "Q Key", "W Key", "E Key", "A Key", "S Key", "D Key", "Z Key", "X Key", "C Key" 
			};



void Engine::LoadData( const std::string& aFile )
{
	Deserialiser lD( aFile, rRootPipeline, this );

	// Serialise data in logic pipelines
	for ( std::multimap<int, grn::Pipeline*>::iterator it = rRootPipeline->iLogicPipelines.begin()
		; it != rRootPipeline->iLogicPipelines.end()
		; it++ )
		lD.DeserialiseInputs(it->second);
		
	// Serialise data in render pipelines
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iRenderPipelines.begin()
		; it != rRootPipeline->iRenderPipelines.end()
		; it++ )
		lD.DeserialiseInputs(*it);
	
	lD.Flush();
}


void Engine::Build( const std::string& aFile )
{
	// Setup and run the game build pipelines
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iBuildPipelines.begin()
		; it != rRootPipeline->iBuildPipelines.end()
		; it++ )
	{
		(*it)->Link( this );
		(*it)->Setup();
		(*it)->Run();
	}
	
	// Now all game data is waiting in various inputs. Save it.
	SerialiseVisitor lSerialiser(aFile);

	// Serialise data in logic pipelines
	for ( std::multimap<int, grn::Pipeline*>::iterator it = rRootPipeline->iLogicPipelines.begin()
		; it != rRootPipeline->iLogicPipelines.end()
		; it++ )
		lSerialiser.SerialiseInputs(it->second);
		
	// Serialise data in render pipelines
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iRenderPipelines.begin()
		; it != rRootPipeline->iRenderPipelines.end()
		; it++ )
		lSerialiser.SerialiseInputs(*it);
	
	lSerialiser.Flush();
	
	rRootPipeline->ClearData();
}


void Engine::Start()
{
	// Setup and run the game build pipelines
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iBuildPipelines.begin()
		; it != rRootPipeline->iBuildPipelines.end()
		; it++ )
	{
		(*it)->Link( this );
		(*it)->Setup();
		(*it)->Run();
	}
	
	// Link input to logic pipelines
	for ( std::multimap<int, grn::Pipeline*>::iterator it = rRootPipeline->iLogicPipelines.begin()
		; it != rRootPipeline->iLogicPipelines.end()
		; it++ )
	{
		it->second->Link( this );
		it->second->Setup();
	}
		
	// Setup the rendering pipeline
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iRenderPipelines.begin()
		; it != rRootPipeline->iRenderPipelines.end()
		; it++ )
	{
		(*it)->Link( this );
		(*it)->Setup();
	}
}



void Engine::Stop()
{
	rRootPipeline->ClearData();
}


void Engine::RunFrame()
{
	// Update enable/disabbled status
	for ( std::multimap<int, grn::Pipeline*>::iterator it = rRootPipeline->iLogicPipelines.begin()
		; it != rRootPipeline->iLogicPipelines.end()
		; it++ )
		it->second->UpdateEnabled();
			
	// Update enable/disabbled status
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iRenderPipelines.begin()
		; it != rRootPipeline->iRenderPipelines.end()
		; it++ )
		(*it)->UpdateEnabled();		

	// Process logic pipelines
	for ( std::multimap<int, grn::Pipeline*>::iterator it = rRootPipeline->iLogicPipelines.begin()
		; it != rRootPipeline->iLogicPipelines.end()
		; it++ )
		it->second->Run();
			
	// Run the render pipelines
	for ( std::list<grn::Pipeline*>::iterator it = rRootPipeline->iRenderPipelines.begin()
		; it != rRootPipeline->iRenderPipelines.end()
		; it++ )
		(*it)->Run();		
}


void Engine::ProcessKey( Engine::KeyCode aCode, bool aPressed )
{
	for ( std::pair< KeyListeners::iterator, KeyListeners::iterator > lRange = iKeyListeners.equal_range(aCode)
		; lRange.first!=lRange.second
		; lRange.first++ )
	{
		lRange.first->second->ProcessKey( aCode, aPressed );
	}
}


