
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#include <GL/glext.h>


#include "RendererGL2.h"

#include "Program.h"
#include "GironaLog.h"


#define NO_TRANSF_PARAMETER 0xFFFFFFF

#include "DepthCode.shaders.h"
#include "UVCode.shaders.h"
#include "Phong.shaders.h"


using namespace grn;


static const int sVertexBufferGLTypes[ Mesh::VertexBuffer::_eNumTypes ] = { GL_FLOAT, GL_UNSIGNED_BYTE	};
static const int sIndexBufferGLTypes[ Mesh::IndexBuffer::_eNumTypes ] = { GL_UNSIGNED_INT, GL_UNSIGNED_SHORT, GL_UNSIGNED_BYTE	};			



RendererGL2::RendererGL2()
{
	iPhongProgram.LoadFromStrings( PIXEL_PHONG_V_SHADER, PIXEL_PHONG_F_SHADER );
	iPhongModelTransformParameter = iPhongProgram.FindParameter("uModelTransform");
	iPhongModelNormalTransformParameter = iPhongProgram.FindParameter("uModelNormalTransform");
	iPhongSpecularParameter = iPhongProgram.FindParameter("uSpecular");
	iPhongDiffuseParameter = iPhongProgram.FindParameter("uDiffuse");
	iPhongAmbientParameter = iPhongProgram.FindParameter("uAmbient");
	iPhongExponentParameter = iPhongProgram.FindParameter("uExponent");
	iPhongEyePosParameter = iPhongProgram.FindParameter("uEyePos");
	iPhongLightPosParameter = iPhongProgram.FindParameter("uLightPos");	
	
	iShadowPhongProgram.LoadFromStrings( SHADOW_PIXEL_PHONG_V_SHADER, SHADOW_PIXEL_PHONG_F_SHADER );
	iShadowPhongDepthBufferParameter = iShadowPhongProgram.FindParameter("uDepthBuffer");
	iShadowPhongDepthTransformParameter = iShadowPhongProgram.FindParameter("uDepthTransform");
	iShadowPhongModelTransformParameter = iShadowPhongProgram.FindParameter("uModelTransform");
	iShadowPhongModelNormalTransformParameter = iShadowPhongProgram.FindParameter("uModelNormalTransform");
	iShadowPhongSpecularParameter = iShadowPhongProgram.FindParameter("uSpecular");
	iShadowPhongDiffuseParameter = iShadowPhongProgram.FindParameter("uDiffuse");
	iShadowPhongAmbientParameter = iShadowPhongProgram.FindParameter("uAmbient");
	iShadowPhongExponentParameter = iShadowPhongProgram.FindParameter("uExponent");
	iShadowPhongEyePosParameter = iShadowPhongProgram.FindParameter("uEyePos");
	iShadowPhongLightPosParameter = iShadowPhongProgram.FindParameter("uLightPos");	
	
	iDepthCodeProgram.LoadFromStrings( DEPTH_V_SHADER, DEPTH_F_SHADER );
	iDepthCodeModelTransformParameter = iDepthCodeProgram.FindParameter("uModelTransform");
	
	iUVCodeProgram.LoadFromStrings( UVCODE_V_SHADER, UVCODE_F_SHADER );
	iUVCodeModelTransformParameter = iUVCodeProgram.FindParameter("uModelTransform");
	
	// Init state
	iNullOutput = false;
	iCurrentModelTransformParameter = NO_TRANSF_PARAMETER;
	iCurrentNormalTransformParameter = NO_TRANSF_PARAMETER;
	rCurrentProgram = NULL;
}


void RendererGL2::Activate( Material* aMat )
{
	glEnable(GL_DEPTH_TEST);

	// Phong material
	if ( PhongMaterial* lPhong = dynamic_cast<PhongMaterial*>(aMat) )
	{
		if (lPhong->iReceiveShadows)
		{
			rCurrentProgram = &iShadowPhongProgram;
			iCurrentModelTransformParameter = iShadowPhongModelTransformParameter;
			iCurrentNormalTransformParameter = iShadowPhongModelNormalTransformParameter;
			
			iShadowPhongProgram.Activate();
			iShadowPhongProgram.SetParameterValue( iShadowPhongSpecularParameter, lPhong->iSpecular.iColor );
			iShadowPhongProgram.SetParameterValue( iShadowPhongDiffuseParameter, lPhong->iDiffuse.iColor );
			iShadowPhongProgram.SetParameterValue( iShadowPhongAmbientParameter, lPhong->iAmbient.iColor );
			iShadowPhongProgram.SetParameterValue( iShadowPhongExponentParameter, lPhong->iExponent );
			iShadowPhongProgram.SetParameterValue( iShadowPhongEyePosParameter, lPhong->iEyePosition );
			iShadowPhongProgram.SetParameterValue( iShadowPhongLightPosParameter, lPhong->iLightPosition );

			iShadowPhongProgram.SetParameterValue( iShadowPhongDepthBufferParameter, 0 );
			if (lPhong->rDepthBuffer) lPhong->rDepthBuffer->Bind();
			iShadowPhongProgram.SetParameterValue( iShadowPhongDepthTransformParameter, lPhong->rDepthMatrix );
			
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			
			static float Far[] = {1.0f,1.0f,1.0f,1.0f};
			glTexParameterfv( GL_TEXTURE_2D, GL_TEXTURE_BORDER_COLOR, Far );
		}
		else
		{
			rCurrentProgram = &iPhongProgram;
			iCurrentModelTransformParameter = iPhongModelTransformParameter;
			iCurrentNormalTransformParameter = iPhongModelNormalTransformParameter;

			iPhongProgram.Activate();
			iPhongProgram.SetParameterValue( iShadowPhongSpecularParameter, lPhong->iSpecular.iColor );
			iPhongProgram.SetParameterValue( iShadowPhongDiffuseParameter, lPhong->iDiffuse.iColor );
			iPhongProgram.SetParameterValue( iShadowPhongAmbientParameter, lPhong->iAmbient.iColor );
			iPhongProgram.SetParameterValue( iShadowPhongExponentParameter, lPhong->iExponent );
			iPhongProgram.SetParameterValue( iShadowPhongEyePosParameter, lPhong->iEyePosition );
			iPhongProgram.SetParameterValue( iShadowPhongLightPosParameter, lPhong->iLightPosition );
		}

		iNullOutput = false;
	}
	
	// Depth code material
	else if ( dynamic_cast<DepthCodeMaterial*>(aMat) )
	{
		rCurrentProgram = &iDepthCodeProgram;
		iCurrentModelTransformParameter = iDepthCodeModelTransformParameter;
		iCurrentNormalTransformParameter = NO_TRANSF_PARAMETER;
		iDepthCodeProgram.Activate();
		iNullOutput = false;
	}
	
	// UV code material
	else if ( dynamic_cast<UVCodeMaterial*>(aMat) )
	{
		rCurrentProgram = &iUVCodeProgram;
		iCurrentModelTransformParameter = iUVCodeModelTransformParameter;
		iCurrentNormalTransformParameter = NO_TRANSF_PARAMETER;
		iUVCodeProgram.Activate();
		iNullOutput = false;
	}
	
	else
	{
		// unsupported material
		iNullOutput = true;
	}
	
}


void RendererGL2::SetModelTransform( const SimdTransform& aT, const SimdTransform& aN)
{
	assert(rCurrentProgram);
	
	if (iCurrentModelTransformParameter!=NO_TRANSF_PARAMETER)
		rCurrentProgram->SetParameterValue( iCurrentModelTransformParameter, aT );
	
	if (iCurrentNormalTransformParameter!=NO_TRANSF_PARAMETER)
		rCurrentProgram->SetParameterValue( iCurrentNormalTransformParameter, aN );
}


void RendererGL2::Deactivate( Material* )
{
	if (rCurrentProgram) rCurrentProgram->Deactivate();
}


void RendererGL2::Activate( const Camera& aCam ) const
{
	float t[16];
	aCam.GetViewMatrix().getValue(t);
	glLoadIdentity();
	glMultMatrixf(t);
	
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	aCam.GetProjectionMatrix().getValue(t);
	glMultMatrixf(t);
	
	glMatrixMode( GL_MODELVIEW );
}



void RendererGL2::Render( const AABB& aB ) const
{
	glBegin(GL_QUADS);
	
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMin.z());
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMin.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMin.z());
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMin.z());
	
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMin.z());
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMin.z());
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMax.z());
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMax.z());
	
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMax.z());
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMax.z());
	
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMin.z());
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMin.z());
	
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMin.z());
	glVertex3f(aB.iMin.x(),aB.iMin.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMax.z());
	glVertex3f(aB.iMax.x(),aB.iMin.y(),aB.iMin.z());
	
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMin.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMin.z());
	glVertex3f(aB.iMax.x(),aB.iMax.y(),aB.iMax.z());
	glVertex3f(aB.iMin.x(),aB.iMax.y(),aB.iMax.z());
	
	glEnd();
}


void RendererGL2::RenderInverted( const AABB& aB ) const
{
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	Render(aB);
	glCullFace(GL_FRONT);
	glDisable(GL_CULL_FACE);
}


void RendererGL2::Batch( Mesh* aMesh )
{
	if (iNullOutput) return;
	
	// Enable buffer states
	glEnableClientState( GL_VERTEX_ARRAY );
	for ( std::vector<Mesh::VertexBuffer*>::iterator it = aMesh->iVertexBuffers.begin()
		; it!=aMesh->iVertexBuffers.end()
		; it++ )
	{
		Mesh::VertexBuffer* lBuffer = *it;

		if (lBuffer->iHandle) glBindBuffer( GL_ARRAY_BUFFER, lBuffer->iHandle );
		
		if (lBuffer->iSemantic==Mesh::VertexBuffer::ePosition)
			glVertexPointer( lBuffer->iNumElements
							, sVertexBufferGLTypes[ lBuffer->iDataType ]
							, 0
							, lBuffer->iHandle ? 0 : lBuffer->iData );
		else
		{
			glEnableVertexAttribArrayARB( lBuffer->iSemantic );
			glVertexAttribPointerARB( lBuffer->iSemantic
									 , lBuffer->iNumElements
									 , sVertexBufferGLTypes[ lBuffer->iDataType ]
									 , 0
									 , 0
									 , lBuffer->iHandle ? 0 : lBuffer->iData );		
		}

		if (lBuffer->iHandle) glBindBuffer( GL_ARRAY_BUFFER, 0 );
		
	} // end for each vertex buffer
	
	// Send
	glDrawElements( GL_TRIANGLES, aMesh->iNumFaces*3, GL_UNSIGNED_SHORT, aMesh->iIndexBuffer->iData );			
	
	// Disable state
	glDisableClientState( GL_VERTEX_ARRAY );	
	for ( std::vector<Mesh::VertexBuffer*>::iterator it = aMesh->iVertexBuffers.begin()
		; it!=aMesh->iVertexBuffers.end()
		; it++ )
		glDisableVertexAttribArrayARB( (*it)->iSemantic );

/* debug version

	float* lBuffer = (float*)(iVertexBuffers[0])->iData;
	float* lNormal = (float*)(iVertexBuffers[1])->iData;
	unsigned short* lIndexBuffer = (unsigned short*)(iIndexBuffer->iData);
	glBegin(GL_TRIANGLES);
	
	for (int i=0;i<iNumFaces; i++)
	{
		glNormal3f( lNormal[ lIndexBuffer[i*3+0]*3+0 ], lNormal[ lIndexBuffer[i*3+0]*3+1 ], lNormal[ lIndexBuffer[i*3+0]*3+2 ] );
		glVertex3f( lBuffer[ lIndexBuffer[i*3+0]*3+0 ], lBuffer[ lIndexBuffer[i*3+0]*3+1 ], lBuffer[ lIndexBuffer[i*3+0]*3+2 ] );
		
		glNormal3f( lNormal[ lIndexBuffer[i*3+1]*3+0 ], lNormal[ lIndexBuffer[i*3+1]*3+1 ], lNormal[ lIndexBuffer[i*3+1]*3+2 ] );
		glVertex3f( lBuffer[ lIndexBuffer[i*3+1]*3+0 ], lBuffer[ lIndexBuffer[i*3+1]*3+1 ], lBuffer[ lIndexBuffer[i*3+1]*3+2 ] );

		glNormal3f( lNormal[ lIndexBuffer[i*3+2]*3+0 ], lNormal[ lIndexBuffer[i*3+2]*3+1 ], lNormal[ lIndexBuffer[i*3+2]*3+2 ] );
		glVertex3f( lBuffer[ lIndexBuffer[i*3+2]*3+0 ], lBuffer[ lIndexBuffer[i*3+2]*3+1 ], lBuffer[ lIndexBuffer[i*3+2]*3+2 ] );
	}
	
	glEnd();

*/

}


void RendererGL2::Upload( Mesh* aMesh, bool aFreeFromMainMemory)
{
	for (unsigned int i=0; i<aMesh->iVertexBuffers.size(); i++)
		UploadBuffer( aMesh->iVertexBuffers[i], aMesh->iNumVertices, aFreeFromMainMemory );
}


void RendererGL2::Remove( Mesh* aMesh )
{
	for (unsigned int i=0; i<aMesh->iVertexBuffers.size(); i++)
		RemoveBuffer( aMesh->iVertexBuffers[i] );
}


void RendererGL2::UploadBuffer( Mesh::VertexBuffer* aBuffer, int aNumVertices, bool aFreeFromMainMemory )
{
	if (aBuffer->iData)
	{
		if (aBuffer->iHandle>0) RemoveBuffer( aBuffer );
		
		glGenBuffers( 1, &(aBuffer->iHandle) );
		glBindBuffer( GL_ARRAY_BUFFER, aBuffer->iHandle );
		glBufferData( GL_ARRAY_BUFFER
					, aNumVertices * aBuffer->iNumElements * Mesh::VertexBuffer::sTypeSizes[aBuffer->iDataType]
					, aBuffer->iData
					, GL_STATIC_DRAW );
		glBindBuffer( GL_ARRAY_BUFFER, 0 );
		
		if (aFreeFromMainMemory) 
		{
			delete[] aBuffer->iData;
			aBuffer->iData = NULL;
		}
	}
}


void RendererGL2::RemoveBuffer( Mesh::VertexBuffer* aBuffer )
{
	if (aBuffer->iHandle>0)
	{
		glDeleteBuffers( 1, &(aBuffer->iHandle) );
		aBuffer->iHandle = 0;
	}			
}


void RendererGL2::RenderScreenQuad( float x0, float y0, float x1, float y1, Texture* aT )
{
	Texture* lT = dynamic_cast<Texture*>(aT);
	if (lT)
	{	
		glMatrixMode( GL_PROJECTION );
		glLoadIdentity();
		glOrtho(-1.0,1.0,-1.0,1.0,-0.5,0.5);
		glMatrixMode( GL_MODELVIEW);
		glLoadIdentity();

		glDisable( GL_DEPTH_TEST );
		glDisable( GL_CULL_FACE );
		glDisable( GL_LIGHTING );

		//lT->iTexture->EnableTextureTarget();
		//lT->iTexture->Bind();
		lT->Bind();

		glBegin( GL_QUADS );
		
		glColor4f( 1.0f, 1.0f, 1.0f, 1.0f );
		
		glTexCoord2f( 0.0f, 0.0f );
		glVertex2f( x0, y0 );
		
		glTexCoord2f( 1.0f, 0.0f );
		glVertex2f( x1, y0 );

		glTexCoord2f( 1.0f, 1.0f );
		glVertex2f( x1, y1 );

		glTexCoord2f( 0.0f, 1.0f );
		glVertex2f( x0, y1 );

		glEnd();			

		lT->Unbind();
		//lT->iTexture->DisableTextureTarget();

		glEnable( GL_LIGHTING );
		glEnable( GL_CULL_FACE );	
		glEnable( GL_DEPTH_TEST );
	}
}


void RendererGL2::RenderScreenQuad( float x0, float y0, float x1, float y1, SimdVector4 aC )
{
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glOrtho(-1.0,1.0,-1.0,1.0,-0.5,0.5);
	glMatrixMode( GL_MODELVIEW);
	glLoadIdentity();

	glDisable( GL_DEPTH_TEST );
	glDisable( GL_CULL_FACE );
	glDisable( GL_LIGHTING );
	glEnable(GL_BLEND);										// Turn Blending On
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);		// Blending Function For Translucency Based On Source Alpha Value

	glBegin( GL_QUADS );
	
	glColor4f( aC[0], aC[1], aC[2], aC[3] );
	
	glVertex2f( x0, y0 );
	glVertex2f( x1, y0 );
	glVertex2f( x1, y1 );
	glVertex2f( x0, y1 );

	glEnd();			

	glDisable(GL_BLEND);		// Turn Blending Off
	glEnable( GL_LIGHTING );
	glEnable( GL_CULL_FACE );	
	glEnable( GL_DEPTH_TEST );
}



ScreenRenderTarget* RendererGL2::NewScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine )
{
	return new GL2ScreenRenderTarget(aX,aY,aWidth,aHeight,aEngine);
}
		

OffscreenRenderTarget* RendererGL2::NewOffscreenRenderTarget( int aWidth, int aHeight, bool aColor=true, bool aDepth=true )
{
	return new GL2OffscreenRenderTarget(aWidth,aHeight,aColor,aDepth);
}


//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////



GL2OffscreenRenderTarget::GL2OffscreenRenderTarget( int aWidth, int aHeight, bool aColor, bool aDepth )
	: OffscreenRenderTarget(aWidth,aHeight)
{


	glGenFramebuffersEXT( 1, &iFrameBuffer );
	glGenRenderbuffersEXT( 1, &iDepthRenderBuffer );
	glGenTextures( 1, &iTex );
	glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, iFrameBuffer );
	
	glBindTexture( GL_TEXTURE_2D, iTex );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA8, aWidth, aHeight, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    
    glFramebufferTexture2DEXT( GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, iTex, 0 );
    
    glBindRenderbufferEXT( GL_RENDERBUFFER_EXT, iDepthRenderBuffer );
    glRenderbufferStorageEXT( GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, aWidth, aHeight );
    
    glFramebufferRenderbufferEXT( GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, iDepthRenderBuffer );
   
	{
		GLenum lStatus;
		lStatus = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		switch(lStatus)
		{
			case GL_FRAMEBUFFER_COMPLETE_EXT: break;
			case GL_FRAMEBUFFER_UNSUPPORTED_EXT: grn::force::log("[error] Unsupported Framebuffer setup."); break;
			default : grn::force::log("[error] Invalid Framebuffer setup.");
		}
	}
    
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
    
}


void GL2OffscreenRenderTarget::Activate()
{
	log("activated offscreen render target.");
//	iTexture->BeginCapture();
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, iFrameBuffer );
	glViewport( 0, 0, iWidth, iHeight );
	glDisable(GL_SCISSOR_TEST);
}


void GL2OffscreenRenderTarget::Deactivate()
{
	glEnable(GL_SCISSOR_TEST);
    glBindFramebufferEXT( GL_FRAMEBUFFER_EXT, 0 );
//	iTexture->EndCapture();
	log("deactivated offscreen render target.");
}


void GL2OffscreenRenderTarget::Present()
{
}


void GL2OffscreenRenderTarget::Clear()
{
	Activate();
	glClearColor(1.0,1.0,1.0,1.0f);
	glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
	Deactivate();
}


Texture* GL2OffscreenRenderTarget::GetColorTexture()
{
	GL2Texture* lT = new GL2Texture();
	//lT->iTextureID = iTexture->GetTextureID();
	//lT->iTextureTarget = iTexture->GetTextureTarget();
	lT->iTextureID = iTex;
	lT->iTextureTarget = GL_TEXTURE_2D;
	return lT;
}


Texture* GL2OffscreenRenderTarget::GetDepthTexture()
{
	GL2Texture* lT = new GL2Texture();
//	lT->iTextureID = iTexture->GetDepthTextureID();
//	lT->iTextureTarget = iTexture->GetTextureTarget();
	lT->iTextureID = iTex;
	lT->iTextureTarget = GL_TEXTURE_2D;
	return lT;
}


//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////




GL2ScreenRenderTarget::GL2ScreenRenderTarget( int aX, int aY, int aWidth, int aHeight, Engine* aEngine )
	: ScreenRenderTarget(aX,aY,aWidth,aHeight,aEngine)
{
}


void GL2ScreenRenderTarget::Activate()
{
	int iSize = SimdMin(iWidth,iHeight);
	glViewport( iX+(iWidth-iSize)/2
			  , iY+(iHeight-iSize)/2
			  , iSize
			  , iSize );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );
	glFrontFace( GL_CCW );
}


void GL2ScreenRenderTarget::Deactivate()
{
}


void GL2ScreenRenderTarget::Clear()
{
	Activate();
	glClearColor(0.0,0.0,0.0,1.0f);
	glClear( GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT );
	Deactivate();
}


//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////
//////////////////////////////////////////////////////


void GL2Texture::Bind()
{
	glEnable( iTextureTarget );
    glBindTexture( iTextureTarget, iTextureID );
}


void GL2Texture::Unbind()
{
	glDisable(iTextureTarget);
}




