
#include "MeshConvert.h"

#include "Pipeline.h"

#include <vector>

using namespace grn;

IMPLEMENT_MODULE(MeshConvert);


MeshConvert::MeshConvert( Pipeline* aParent )
: Module(aParent)
{
	setSize(100,50);
	iX = 250;
	rInput = new Module::tInputPad<grn::BlenderMesh>( "InputMeshes", this );
	rOutput = new Module::tOutputPad<grn::Mesh>( "OutputMeshes", this );
}


class VertexData
{
public:
	SimdPoint3 iPosition;
	SimdPoint3 iNormal;
	SimdPoint3 iTex;
	
	VertexData()
	{
		iPosition = SimdVector3(0.0,0.0,0.0);
		iNormal = SimdVector3(0.0,0.0,0.0);
		iTex = SimdVector3(0.0,0.0,0.0);
	}
	
	bool operator == ( const VertexData& aOther ) const
	{
		if ((iPosition-aOther.iPosition).length2()>1e-5) return false;
		if ((iNormal-aOther.iNormal).length2()>1e-5) return false;
		if ((iTex-aOther.iTex).length2()>1e-5) return false;
		return true;
	}
};


class FaceData
{
public:
	int iIndices[3];
};


SimdVector3 CalculateBlenderFaceNormal( BlenderMesh* lInMesh, int i1, int i2, int i3 )
{
	SimdVector3 p1( lInMesh->rMesh->mvert[i1].co[0], lInMesh->rMesh->mvert[i1].co[1], lInMesh->rMesh->mvert[i1].co[2] );
	SimdVector3 p2( lInMesh->rMesh->mvert[i2].co[0], lInMesh->rMesh->mvert[i2].co[1], lInMesh->rMesh->mvert[i2].co[2] );
	SimdVector3 p3( lInMesh->rMesh->mvert[i3].co[0], lInMesh->rMesh->mvert[i3].co[1], lInMesh->rMesh->mvert[i3].co[2] );
	SimdVector3 n = cross( p2-p1, p3-p1 ).safe_normalized();
	return n;
}


void AddFaceToRawData( BlenderMesh* lInMesh, std::vector<VertexData>& lVertices, std::vector<FaceData>& lFaces
					 , MFace* lFace, TFace* lTexFace
					 , int i1, int i2, int i3
					 , int fi1, int fi2, int fi3 )
{
	SimdVector3 lFaceNormal = CalculateBlenderFaceNormal( lInMesh, i1, i2, i3 );
	
	FaceData lF;
	
	for (int v=0; v<3; v++)
	{
		int lIndex = ( v==0?i1:(v==1?i2:i3) );
		int lFIndex = ( v==0?fi1:(v==1?fi2:fi3) );
		VertexData lV;
		
		// position
		lV.iPosition[0] = lInMesh->rMesh->mvert[lIndex].co[0];
		lV.iPosition[1] = lInMesh->rMesh->mvert[lIndex].co[1];
		lV.iPosition[2] = lInMesh->rMesh->mvert[lIndex].co[2];

		// normal
		if ( lFace->flag & ME_SMOOTH )
		{
			lV.iNormal[0] = lInMesh->rMesh->mvert[lIndex].no[0];
			lV.iNormal[1] = lInMesh->rMesh->mvert[lIndex].no[1];
			lV.iNormal[2] = lInMesh->rMesh->mvert[lIndex].no[2];
		}
		else lV.iNormal = lFaceNormal;
		
		// tex coords
		if ( (lTexFace) )
		{
			lV.iTex[0] = lTexFace->uv[lFIndex][0];
			lV.iTex[1] = lTexFace->uv[lFIndex][1];
			lV.iTex[2] = 0.0f;
		}
		else lV.iTex = SimdVector3(0.0f,0.0f,0.0f);
		
		// add unique
		bool lFound = false;
		for ( unsigned int f=0; (f<lVertices.size()) && (!lFound) ; f++ )
			if ( lVertices[f]==lV )
			{
				lFound = true;
				lF.iIndices[v] = f;
			}
		if (!lFound)
		{
			lF.iIndices[v] = lVertices.size();
			lVertices.push_back(lV);
		}
	}
	
	lFaces.push_back(lF);
}


void MeshConvert::Run()
{
	while (rInput->hasData())
	{
		Pointer<BlenderMesh> lInMesh = rInput->getData(true);

		// Create output mesh		
		Mesh* lOutMesh = new Mesh();

		// Generate raw data		
		std::vector<VertexData> lVertices;
		std::vector<FaceData> lFaces;

		MFace* lFace = lInMesh->rMesh->mface;
		TFace* lTFace = lInMesh->rMesh->tface;
		for (int i=0; i<lInMesh->rMesh->totface; i++)
		{
			if ( (lFace->v4 == 0) && (lFace->v3 > 0) )
			{
				AddFaceToRawData( lInMesh, lVertices, lFaces, lFace, lTFace, lFace->v1,lFace->v2,lFace->v3,0,1,2);
			}
			else if (lFace->v4 > 0)
			{
				AddFaceToRawData( lInMesh, lVertices, lFaces, lFace, lTFace, lFace->v1,lFace->v2,lFace->v3,0,1,2);
				AddFaceToRawData( lInMesh, lVertices, lFaces, lFace, lTFace, lFace->v1,lFace->v3,lFace->v4,0,2,3);
			}	
			lFace++;
			if (lTFace) lTFace++;
		}

		// Create vertex buffers
		lOutMesh->iNumVertices = lVertices.size();
		Mesh::VertexBuffer* lPosBuf = new Mesh::VertexBuffer();
		lPosBuf->iHandle = 0;
		lPosBuf->iSemantic = Mesh::VertexBuffer::ePosition;
		lPosBuf->iDataType = Mesh::VertexBuffer::eFloat;
		lPosBuf->iNumElements = 3;
		lPosBuf->iSize = lPosBuf->iNumElements * lOutMesh->iNumVertices * sizeof(float);
		lPosBuf->iData = (char*) new float[ lPosBuf->iNumElements * lOutMesh->iNumVertices ];
		lOutMesh->iVertexBuffers.push_back(lPosBuf);

		Mesh::VertexBuffer* lNorBuf = new Mesh::VertexBuffer();
		lNorBuf->iHandle = 0;
		lNorBuf->iSemantic = Mesh::VertexBuffer::eNormal;
		lNorBuf->iDataType = Mesh::VertexBuffer::eFloat;
		lNorBuf->iNumElements = 3;
		lNorBuf->iSize = lNorBuf->iNumElements * lOutMesh->iNumVertices * sizeof(float);
		lNorBuf->iData = (char*) new float[ lNorBuf->iNumElements * lOutMesh->iNumVertices ];
		lOutMesh->iVertexBuffers.push_back(lNorBuf);

		Mesh::VertexBuffer* lTexBuf = 0;
		if (lTFace)
		{
			lTexBuf = new Mesh::VertexBuffer();
			lTexBuf->iHandle = 0;
			lTexBuf->iSemantic = Mesh::VertexBuffer::eTexCoord0;
			lTexBuf->iDataType = Mesh::VertexBuffer::eFloat;
			lTexBuf->iNumElements = 2;
			lTexBuf->iSize = lTexBuf->iNumElements * lOutMesh->iNumVertices * sizeof(float);
			lTexBuf->iData = (char*) new float[ lTexBuf->iNumElements * lOutMesh->iNumVertices ];
			lOutMesh->iVertexBuffers.push_back(lTexBuf);
		}

		// Fill buffers		
		float* lData = (float*)lPosBuf->iData;
		float* lDataNormal = (float*)lNorBuf->iData;
		float* lDataTex = lTexBuf ? (float*)lTexBuf->iData : 0;
		for (unsigned int i=0; i<lVertices.size(); i++)
		{
			lData[3*i+0] = lVertices[i].iPosition[0];
			lData[3*i+1] = lVertices[i].iPosition[1];
			lData[3*i+2] = lVertices[i].iPosition[2];
			lOutMesh->iAABB.Bound( SimdVector3(lData[3*i+0],lData[3*i+1],lData[3*i+2]) );

			lDataNormal[3*i+0] = lVertices[i].iNormal[0];
			lDataNormal[3*i+1] = lVertices[i].iNormal[1];
			lDataNormal[3*i+2] = lVertices[i].iNormal[2];
			
			if (lDataTex)
			{
				lDataTex[2*i+0] = lVertices[i].iTex[0];
				lDataTex[2*i+1] = lVertices[i].iTex[1];
			}
		}
				
		// Convert faces
		lOutMesh->iNumFaces = lFaces.size();
		lOutMesh->iIndexBuffer = new Mesh::IndexBuffer();
		lOutMesh->iIndexBuffer->iHandle = 0;
		lOutMesh->iIndexBuffer->iDataType = Mesh::IndexBuffer::eShort;
		lOutMesh->iIndexBuffer->iSize = lOutMesh->iNumFaces * 3 * sizeof(short);
		lOutMesh->iIndexBuffer->iData = (char*) new unsigned short[ lOutMesh->iNumFaces * 3 ];
		unsigned short* lIData = (unsigned short*) lOutMesh->iIndexBuffer->iData;

		// Fill index array
		for (unsigned int i=0; i<lFaces.size(); i++)
		{
			lIData[i*3+0] = lFaces[i].iIndices[0];
			lIData[i*3+1] = lFaces[i].iIndices[1];
			lIData[i*3+2] = lFaces[i].iIndices[2];
		}
		
		rPipeline->Emit(rOutput,lOutMesh);
	}
}
